"""Interface adapter to lifelines Cox PH surival model."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["fkiraly"]

from skpro.survival.adapters.lifelines import _LifelinesAdapter
from skpro.survival.base import BaseSurvReg


class CoxPHlifelines(_LifelinesAdapter, BaseSurvReg):
    r"""Cox proportional hazards models, from lifelines.

    Direct interface to ``lifelines.fitters.CoxPHFitter``,
    by ``CamDavidsonPilon``.

    This class implements Cox proportional hazard model,

    .. math::  h(t|x) = h_0(t) \exp((x - \overline{x})' \beta)

    with different options to fit the baseline hazard, :math:`h_0(t)`.

    The class offers multiple options via the ``baseline_estimation_method`` parameter:

    ``"breslow"`` (default): non-parametric estimate via Breslow's method.
    In this case, the entire model is the traditional semi-parametric Cox model.
    Ties are handled using Efron's method.

    ``"spline"``: parametric spline fit of baseline hazard,
    via Royston-Parmar's method [1]_. The parametric form is

    .. math:: H_0(t) = \exp{\left( \phi_0 + \phi_1\log{t} + \sum_{j=2}^N \phi_j v_j(\log{t})\right)}  # noqa E501

    where :math:`v_j` are our cubic basis functions at predetermined knots,
    and :math:`H_0` is the cumulative baseline hazard. See [1]_ for exact definition.

    ``"piecewise"``: non-parametric, piecewise constant empirical baseline hazard.
    The explicit form of the baseline hazard is

    .. math::  h_0(t) = \begin{cases}
        exp{\beta \cdot \text{center}(x)}  & \text{if $t \le \tau_0$} \\
        exp{\beta \cdot \text{center}(x)} \cdot lambda_1 & \text{if $\tau_0 < t \le \tau_1$} \\  # noqa E501
        exp{\beta \cdot \text{center}(x)} \cdot lambda_2 & \text{if $\tau_1 < t \le \tau_2$} \\  # noqa E501
        ...
        \end{cases}

    Parameters
    ----------
    alpha: float, optional (default=0.05)
      the level in the confidence intervals around the estimated survival function,
      for computation of ``confidence_intervals_`` fitted parameter.

    baseline_estimation_method: string, default="breslow",
        one of: ``"breslow"``, ``"spline"``, or ``"piecewise"``.
        Specifies algorithm for estimation of baseline hazard, see above.
        If ``"piecewise"``, the ``breakpoints`` parameter must be set.

    penalizer: float or array, optional (default=0.0)
        Penalty to the size of the coefficients during regression.
        This improves stability of the estimates and controls for high correlation
        between covariates.
        For example, this shrinks the magnitude value of :math:`\beta_i`.
        See ``l1_ratio`` below.
        The penalty term is :math:`\text{penalizer} \left( \frac{1-\text{l1_ratio}}{2} ||\beta||_2^2 + \text{l1_ratio}||\beta||_1\right)`.  # noqa E501

        If an array, must be equal in size to the number of parameters,
        with penalty coefficients for specific variables. For
        example, ``penalizer=0.01 * np.ones(p)`` is the same as ``penalizer=0.01``.

    l1_ratio: float, optional (default=0.0)
        Specify what ratio to assign to a L1 vs L2 penalty.
        Same as in scikit-learn. See ``penalizer`` above.

    strata: list, optional
        specify a list of columns to use in stratification. This is useful if a
        categorical covariate does not obey the proportional hazard assumption. This
        is used similar to the ``strata`` expression in R.
        See http://courses.washington.edu/b515/l17.pdf.

    n_baseline_knots: int, optional, default=4
        Used only when ``baseline_estimation_method="spline"``.
        Set the number of knots (interior & exterior) in the baseline hazard,
        which will be placed evenly along the time axis.
        Should be at least 2.
        Royston et. al, the authors of this model, suggest 4 to start,
        but any values between 2 and 8 are reasonable.
        If you need to customize the timestamps used to calculate the curve,
        use the ``knots`` parameter instead.

    knots: list, optional
        Used only when ``baseline_estimation_method="spline"``.
        Specifies custom points in the time axis for the baseline hazard curve.
        To use evenly-spaced points in time, the ``n_baseline_knots``
        parameter can be employed instead.

    breakpoints: list, optional
        Used only when ``baseline_estimation_method="piecewise"``,
        must be passed in this case.
        Set the positions of the baseline hazard breakpoints.

    Attributes
    ----------
    params_ : Series
        The estimated coefficients.
    hazard_ratios_ : Series
        The exp(coefficients)
    confidence_intervals_ : DataFrame
        The lower and upper confidence intervals for the hazard coefficients
    durations: Series
        The durations provided
    event_observed: Series
        The event_observed variable provided
    weights: Series
        The event_observed variable provided
    variance_matrix_ : DataFrame
        The variance matrix of the coefficients
    strata: list
        the strata provided
    standard_errors_: Series
        the standard errors of the estimates
    log_likelihood_: float
        the log-likelihood at the fitted coefficients
    AIC_: float
        the AIC at the fitted coefficients (if using splines for baseline hazard)
    partial_AIC_: float
        the AIC at the fitted coefficients
        (if using non-parametric inference for baseline hazard)
    baseline_hazard_: DataFrame
        the baseline hazard evaluated at the observed times.
        Estimated using Breslow's method.
    baseline_cumulative_hazard_: DataFrame
        the baseline cumulative hazard evaluated at the observed times.
        Estimated using Breslow's method.
    baseline_survival_: DataFrame
        the baseline survival evaluated at the observed times.
        Estimated using Breslow's method.
    summary: Dataframe
        a Dataframe of the coefficients, p-values, CIs, etc.

    References
    ----------
    .. [1] Royston, P., Parmar, M. K. B. (2002).
      Flexible parametric proportional-hazards and proportional-odds
      models for censored survival data, with application to prognostic
      modelling and estimation of treatment effects.
      Statistics in Medicine, 21(15), 2175–2197. doi:10.1002/sim.1203
    """

    _tags = {"authors": ["CamDavidsonPilon", "JoseLlanes", "mathurinm", "fkiraly"]}
    # CamDavidsonPilon, JoseLlanes, mathurinm credit for interfaced estimator

    def __init__(
        self,
        baseline_estimation_method: str = "breslow",
        penalizer=0.0,
        strata=None,
        l1_ratio=0.0,
        n_baseline_knots=None,
        knots=None,
        breakpoints=None,
    ):
        self.baseline_estimation_method = baseline_estimation_method
        self.penalizer = penalizer
        self.strata = strata
        self.l1_ratio = l1_ratio
        self.n_baseline_knots = n_baseline_knots
        self.knots = knots
        self.breakpoints = breakpoints

        super().__init__()

    def _get_lifelines_class(self):
        """Getter of the lifelines class to be used for the adapter."""
        from lifelines.fitters.coxph_fitter import CoxPHFitter

        return CoxPHFitter

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params1 = {}

        params2 = {
            "baseline_estimation_method": "spline",
            "penalizer": 0.1,
            "l1_ratio": 0.1,
            "n_baseline_knots": 3,
        }

        # breakpoints are specific to data ranges,
        # but tests loop over various data sets, so this would break
        #
        # params3 = {
        #     "baseline_estimation_method": "piecewise",
        #     "penalizer": 0.15,
        #     "l1_ratio": 0.05,
        #     "breakpoints": [10, 20, 30, 100],
        # }

        return [params1, params2]



"""Base classes for probabilistic metrics."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
# adapted from sktime

from warnings import warn

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype
from sklearn.utils import check_array, check_consistent_length

from skpro.base import BaseObject
from skpro.datatypes import check_is_scitype, convert, convert_to
from skpro.metrics._coerce import _coerce_to_df, _coerce_to_scalar

__author__ = ["fkiraly", "euanenticott-shell"]


class BaseProbaMetric(BaseObject):
    """Base class for probabilistic supervised error metrics in skpro.

    Parameters
    ----------
    multioutput : {'raw_values', 'uniform_average'} or array-like of shape \
            (n_outputs,), default='uniform_average'
        Defines whether and how to aggregate metric for across variables.

        * If 'uniform_average' (default), errors are mean-averaged across variables.
        * If array-like, errors are weighted averaged across variables,
          values as weights.
        * If 'raw_values', does not average errors across variables,
          columns are retained.

    score_average : bool, optional, default=True
        for interval and quantile losses only

        * if True, metric/loss is averaged by upper/lower and/or quantile
        * if False, metric/loss is not averaged by upper/lower and/or quantile
    """

    _tags = {
        "object_type": "metric",  # type of object
        "reserved_params": ["multioutput", "score_average"],
        "scitype:y_pred": "pred_proba",
        "lower_is_better": True,
    }

    def __init__(self, multioutput="uniform_average", score_average=True):
        self.multioutput = multioutput
        self.score_average = score_average
        super().__init__()

    def __call__(self, y_true, y_pred, **kwargs):
        """Calculate metric value using underlying metric function.

        Parameters
        ----------
        y_true : pd.Series, pd.DataFrame, 1D np.array, or 2D np.ndarray
            Ground truth (correct) target values.

        y_pred : return object of probabilistic predictition method scitype:y_pred
            must have same index and columns as y_true
            Predicted values, i-th row is prediction for i-th row of ``y_true``.

        Returns
        -------
        loss : float or 1-column pd.DataFrame with calculated metric value(s)
            metric is always averaged (arithmetic) over instances

            * if ``multioutput = "raw_values"``,
              will have a column level corresponding to variables in ``y_true``
            * if ``multioutput = multioutput = "uniform_average"`` or or array-like
              entries will be averaged over output variable column
            * if ``score_average = False``,
              will have column levels corresponding to quantiles/intervals
            * if ``score_average = True``,
              entries will be averaged over quantiles/interval column
        """
        return self.evaluate(y_true, y_pred, **kwargs)

    def evaluate(self, y_true, y_pred, **kwargs):
        """Evaluate the metric on given inputs.

        Parameters
        ----------
        y_true : pd.Series, pd.DataFrame, 1D np.array, or 2D np.ndarray
            Ground truth (correct) target values.

        y_pred : return object of probabilistic predictition method scitype:y_pred
            must have same index and columns as y_true
            Predicted values, i-th row is prediction for i-th row of ``y_true``.

        Returns
        -------
        loss : float or 1-column pd.DataFrame with calculated metric value(s)
            metric is always averaged (arithmetic) over instances

            * if ``multioutput = "raw_values"``,
              will have a column level corresponding to variables in ``y_true``
            * if ``multioutput = multioutput = "uniform_average"`` or or array-like
              entries will be averaged over output variable column
            * if ``score_average = False``,
              will have column levels corresponding to quantiles/intervals
            * if ``score_average = True``,
              entries will be averaged over quantiles/interval column
        """
        # Input checks and conversions
        y_true_inner, y_pred_inner, multioutput = self._check_ys(
            y_true, y_pred, self.multioutput
        )

        # Don't want to include scores for 0 width intervals, makes no sense
        if 0 in y_pred_inner.columns.get_level_values(1):
            y_pred_inner = y_pred_inner.drop(0, axis=1, level=1)
            warn(
                "Dropping 0 width interval, don't include 0.5 quantile"
                "for interval metrics."
            )

        # pass to inner function
        out = self._evaluate(y_true_inner, y_pred_inner, **kwargs)

        if isinstance(multioutput, str):
            if self.score_average and multioutput == "uniform_average":
                out = out.mean(axis=1).iloc[0]  # average over all
            if self.score_average and multioutput == "raw_values":
                out = out.T.groupby(level=0).mean().T  # average over scores
            if not self.score_average and multioutput == "uniform_average":
                out = out.T.groupby(level=1).mean().T  # average over variables
            if not self.score_average and multioutput == "raw_values":
                out = out  # don't average
        else:  # is np.array with weights
            if self.score_average:
                out_raw = out.T.groupby(level=0).mean().T
                out = out_raw.dot(multioutput)[0]
            else:
                out = _groupby_dot(out, multioutput)

        if isinstance(out, pd.DataFrame):
            out = out.squeeze(axis=0)

        return out

    def _evaluate(self, y_true, y_pred, **kwargs):
        """Evaluate the metric on given inputs.

        Private _evaluate, called by public evaluate.

        Parameters
        ----------
        y_true : pd.Series, pd.DataFrame, 1D np.array, or 2D np.ndarray
            Ground truth (correct) target values.

        y_pred : return object of probabilistic predictition method scitype:y_pred
            must have same index and columns as y_true
            Predicted values, i-th row is prediction for i-th row of ``y_true``.

        Returns
        -------
        loss : pd.DataFrame of shape (, n_outputs), calculated loss metric.
        """
        # Default implementation relies on implementation of evaluate_by_index
        try:
            index_df = self._evaluate_by_index(y_true, y_pred)
            out_df = pd.DataFrame(index_df.mean(axis=0)).T
            out_df.columns = index_df.columns
            return out_df
        except RecursionError as _err:
            msg = (
                f"{type(self).__name__} must implement one of"
                "_evaluate or _evaluate_by_index, but none of the two was implemented"
            )
            raise RecursionError(msg) from _err

    def evaluate_by_index(self, y_true, y_pred, **kwargs):
        """Evaluate the metric by instance index (row).

        Parameters
        ----------
        y_true : pd.Series, pd.DataFrame, 1D np.array, or 2D np.ndarray
            Ground truth (correct) target values.

        y_pred : return object of probabilistic predictition method scitype:y_pred
            must have same index and columns as y_true
            Predicted values, i-th row is prediction for i-th row of ``y_true``.

        Returns
        -------
        loss : pd.DataFrame of same length as ``y_true``,
            calculated metric value(s).

            i-th column contains metric value(s) for prediction at i-th fh element

            * if multioutput = "raw_values",
              will have a column level corresponding to variables in y_true
            * if multioutput = multioutput = "uniform_average" or or array-like
              entries will be averaged over output variable column
            * if score_average = False,
              will have column levels corresponding to quantiles/intervals
            * if score_average = True,
              entries will be averaged over quantiles/interval column
        """
        # Input checks and conversions
        y_true_inner, y_pred_inner, multioutput = self._check_ys(
            y_true, y_pred, self.multioutput
        )

        # Don't want to include scores for 0 width intervals, makes no sense
        if 0 in y_pred_inner.columns.get_level_values(1):
            y_pred_inner = y_pred_inner.drop(0, axis=1, level=1)
            warn(
                "Dropping 0 width interval, don't include 0.5 quantile"
                "for interval metrics."
            )

        # pass to inner function
        out = self._evaluate_by_index(y_true_inner, y_pred_inner, **kwargs)

        if isinstance(multioutput, str):
            if self.score_average and multioutput == "uniform_average":
                out = out.mean(axis=1)  # average over all
            if self.score_average and multioutput == "raw_values":
                out = out.T.groupby(level=0).mean().T  # average over scores
            if not self.score_average and multioutput == "uniform_average":
                out = out.T.groupby(level=1).mean().T  # average over variables
            if not self.score_average and multioutput == "raw_values":
                out = out  # don't average
        else:  # is np.array with weights
            if self.score_average:
                out_raw = out.T.groupby(level=0).mean().T
                out = out_raw.dot(multioutput)[0]
            else:
                out = _groupby_dot(out, multioutput)

        return out

    def _evaluate_by_index(self, y_true, y_pred, **kwargs):
        """Evaluate the metric by instance index (row).

        Private _evaluate_by_index, called by public evaluate_by_index.

        By default this uses _evaluate to find jackknifed pseudosamples. This
        estimates the error at each of the time points.

        Parameters
        ----------
        y_true : pd.Series, pd.DataFrame, 1D np.array, or 2D np.ndarray
            Ground truth (correct) target values.

        y_pred : return object of probabilistic predictition method scitype:y_pred
            must have same index and columns as ``y_true``.
            Predicted values, i-th row is prediction for i-th row of ``y_true``.
        """
        n = y_true.shape[0]
        out_series = pd.Series(index=y_pred.index)
        try:
            x_bar = self.evaluate(y_true, y_pred, self.multioutput, **kwargs)
            for i in range(n):
                out_series[i] = n * x_bar - (n - 1) * self.evaluate(
                    np.vstack((y_true[:i, :], y_true[i + 1 :, :])),  # noqa
                    np.vstack((y_pred[:i, :], y_pred[i + 1 :, :])),  # noqa
                    self.multioutput,
                    **kwargs,
                )
            return out_series
        except RecursionError:
            raise RecursionError(
                "Must implement one of _evaluate or _evaluate_by_index"
            )

    def _check_consistent_input(self, y_true, y_pred, multioutput):
        check_consistent_length(y_true, y_pred)

        y_true = check_array(y_true, ensure_2d=False)

        if not isinstance(y_pred, pd.DataFrame):
            raise ValueError("y_pred should be a dataframe.")

        if not np.all([is_numeric_dtype(y_pred[c]) for c in y_pred.columns]):
            raise ValueError("Data should be numeric.")

        if y_true.ndim == 1:
            y_true = y_true.reshape((-1, 1))

        n_outputs = y_true.shape[1]

        allowed_multioutput_str = ("raw_values", "uniform_average", "variance_weighted")
        if isinstance(multioutput, str):
            if multioutput not in allowed_multioutput_str:
                raise ValueError(
                    "Allowed 'multioutput' string values are {}. "
                    "You provided multioutput={!r}".format(
                        allowed_multioutput_str, multioutput
                    )
                )
        elif multioutput is not None:
            multioutput = check_array(multioutput, ensure_2d=False)
            if n_outputs == 1:
                raise ValueError("Custom weights are useful only in multi-output case.")
            elif n_outputs != len(multioutput):
                raise ValueError(
                    "There must be equally many custom weights (%d) as outputs (%d)."
                    % (len(multioutput), n_outputs)
                )

        return y_true, y_pred, multioutput

    def _check_ys(self, y_true, y_pred, multioutput):
        if multioutput is None:
            multioutput = self.multioutput

        valid, msg, metadata = check_is_scitype(
            y_pred, scitype="Proba", return_metadata=True, var_name="y_pred"
        )

        if not valid:
            raise TypeError(msg)

        y_pred_mtype = metadata["mtype"]
        inner_y_pred_mtype = self.get_tag("scitype:y_pred")

        y_pred_inner = convert(
            y_pred,
            from_type=y_pred_mtype,
            to_type=inner_y_pred_mtype,
            as_scitype="Proba",
        )

        if inner_y_pred_mtype == "pred_interval":
            if 0.0 in y_pred_inner.columns.get_level_values(1):
                for var in y_pred_inner.columns.get_level_values(0):
                    y_pred_inner[var, 0.0, "upper"] = y_pred_inner[var, 0.0, "lower"]

        y_true, y_pred, multioutput = self._check_consistent_input(
            y_true, y_pred, multioutput
        )

        return y_true, y_pred_inner, multioutput

    def _get_alpha_from(self, y_pred):
        """Fetch the alphas present in y_pred."""
        alphas = np.unique(list(y_pred.columns.get_level_values(1)))
        if not all((alphas > 0) & (alphas < 1)):
            raise ValueError("Alpha must be between 0 and 1.")

        return alphas

    def _check_alpha(self, alpha):
        """Check alpha input and coerce to np.ndarray."""
        if alpha is None:
            return None

        if isinstance(alpha, float):
            alpha = [alpha]

        if not isinstance(alpha, np.ndarray):
            alpha = np.asarray(alpha)

        if not all((alpha > 0) & (alpha < 1)):
            raise ValueError("Alpha must be between 0 and 1.")

        return alpha

    def _check_coverage(self, coverage):
        """Check coverage input and coerce to np.ndarray."""
        if coverage is None:
            return None

        if isinstance(coverage, float):
            coverage = [coverage]

        if not isinstance(coverage, np.ndarray):
            coverage = np.asarray(coverage)

        if not all((coverage > 0) & (coverage < 1)):
            raise ValueError("Coverage must be between 0 and 1.")

        return coverage

    def _handle_multioutput(self, loss, multioutput):
        """Handle output according to multioutput parameter.

        Parameters
        ----------
        loss : float, np.ndarray the evaluated metric value.

        multioutput : string "uniform_average" or "raw_values" determines how \
            multioutput results will be treated.
        """
        if isinstance(multioutput, str):
            if multioutput == "raw_values":
                return loss
            elif multioutput == "uniform_average":
                # pass None as weights to np.average: uniform mean
                multioutput = None
            else:
                raise ValueError(
                    "multioutput is expected to be 'raw_values' "
                    "or 'uniform_average' but we got %r"
                    " instead." % multioutput
                )

        if loss.ndim > 1:
            out = np.average(loss, weights=multioutput, axis=1)
        else:
            out = np.average(loss, weights=multioutput)
        return out


class BaseDistrMetric(BaseProbaMetric):
    """Intermediate base class for distributional prediction metrics/scores.

    Developer note:
    Experimental and overrides public methods of BaseProbaMetric.
    This should be refactored into one base class.
    """

    _tags = {
        "object_type": ["metric", "metric_distr"],  # type of object
        "scitype:y_pred": "pred_proba",
        "lower_is_better": True,
    }

    def evaluate(self, y_true, y_pred, **kwargs):
        """Evaluate the  metric on given inputs.

        Parameters
        ----------
        y_true : pd.Series, pd.DataFrame, 1D np.array, or 2D np.ndarray
            Ground truth (correct) target values.

        y_pred : return object of probabilistic predictition method scitype:y_pred
            must have same index and columns as y_true
            Predicted values, i-th row is prediction for i-th row of ``y_true``.

        Returns
        -------
        loss : float or 1-column pd.DataFrame with calculated metric value(s)

            float if multioutput = "uniform_average" or multivariate = True

            1-column df if multioutput = "raw_values" and metric is not multivariate
            metric is always averaged (arithmetic) over rows
        """
        multioutput = self.multioutput
        multivariate = self.multivariate

        index_df = self.evaluate_by_index(y_true, y_pred, **kwargs)
        out_df = pd.DataFrame(index_df.mean(axis=0)).T
        out_df.columns = index_df.columns

        if multioutput == "uniform_average" and not multivariate:
            out_df = out_df.mean(axis=1)
        if multioutput == "uniform_average" or multivariate:
            out = _coerce_to_scalar(out_df)
        else:
            out = _coerce_to_df(out_df)
        return out

    def _coerce_inner_df(self, obj):
        """Coerce obj to pd_DataFrame_Table, for inner method call.

        Parameters
        ----------
        obj : object
            Object to coerce

        Returns
        -------
        obj : object
            Coerced object
        """
        if obj is None:
            return None
        obj = convert_to(obj, to_type="pd_DataFrame_Table", as_scitype="Table")
        obj = _coerce_to_df(obj)
        return obj

    def evaluate_by_index(self, y_true, y_pred, **kwargs):
        """Evaluate the metric by instance index (row).

        Parameters
        ----------
        y_true : pd.Series, pd.DataFrame, 1D np.array, or 2D np.ndarray
            Ground truth (correct) target values.

        y_pred : skpro BaseDistribution of same shape as y_true
            Predictive distribution.
            Must have same index and columns as ``y_true``.

        Returns
        -------
        loss : ``pd.Series`` or ``pd.DataFrame``
            Calculated metric, by time point (default=jackknife pseudo-values).

            ``pd.Series`` if ``self.multioutput="uniform_average"`` or array-like

            * index is equal to index of ``y_true``
            * entry at index i is metric at time i, averaged over variables

            ``pd.DataFrame`` if ``self.multioutput="raw_values"``

            * index and columns equal to those of ``y_true``
            * i,j-th entry is metric at time i, at variable j
        """
        multioutput = self.multioutput

        if hasattr(self, "multivariate"):
            multivariate = self.multivariate
        else:
            multivariate = False

        y_true = self._coerce_inner_df(y_true)

        if "C_true" in kwargs:
            C_true = kwargs["C_true"]
            C_true = self._coerce_inner_df(C_true)
            kwargs_inner = {"C_true": C_true}
        else:
            kwargs_inner = {}

        if multivariate:
            res = self._evaluate_by_index(
                y_true=y_true, y_pred=y_pred, multioutput=multioutput, **kwargs_inner
            )
            res.columns = ["score"]
            return res
        else:
            res_by_col = []
            for col in y_pred.columns:
                y_pred_col = y_pred.loc[:, [col]]
                y_true_col = y_true.loc[:, [col]]
                res_for_col = self._evaluate_by_index(
                    y_true=y_true_col,
                    y_pred=y_pred_col,
                    multioutput=multioutput,
                    **kwargs_inner,
                )
                res_for_col.columns = [col]
                res_by_col += [res_for_col]
            res = pd.concat(res_by_col, axis=1)

        return res


def _groupby_dot(df, weights):
    """Groupby dot product.

    Groups df by axis 1, level 1, and applies dot product with weights.

    Parameters
    ----------
    df : pd.DataFrame
        dataframe to groupby
    weights : np.array
        weights to apply to each group

    Returns
    -------
    out : pd.DataFrame
        dataframe with weighted groupby dot product
    """
    out = df.T.groupby(level=1).apply(lambda x: x.T.dot(weights)).T
    # formerly
    # out = df.groupby(axis=1, level=1).apply(lambda x: x.dot(weights))
    # but groupby(axis=1) is deprecated
    return out


class BaseSurvDistrMetric(BaseDistrMetric):
    """Intermediate base class for distributional prediction metrics/scores.

    Developer note:
    Same as BaseSurvDistrMetric, except for tag set and docstring overrides.
    """

    _tags = {"capability:survival": True}

    def evaluate(self, y_true, y_pred, **kwargs):
        """Evaluate the  metric on given inputs.

        Parameters
        ----------
        y_true : pd.Series, pd.DataFrame, 1D np.array, or 2D np.ndarray
            Ground truth (correct) target values.

        y_pred : return object of probabilistic predictition method scitype:y_pred
            must have same index and columns as y_true
            Predicted values, i-th row is prediction for i-th row of ``y_true``.

        C_true : pd.Series, pd.DataFrame, np.ndarray, optional (default=None)
            censoring information for survival analysis,
            should have same column name as y, same length as X and y
            should have entries 0 and 1 (float or int)
            0 = uncensored, 1 = (right) censored
            if None, all observations are assumed to be uncensored

        Returns
        -------
        loss : float or 1-column pd.DataFrame with calculated metric value(s)
            float if multioutput = "uniform_average" or multivariate = True
            1-column df if multioutput = "raw_values" and metric is not multivariate
            metric is always averaged (arithmetic) over rows
        """
        return super().evaluate(y_true=y_true, y_pred=y_pred, **kwargs)

    def evaluate_by_index(self, y_true, y_pred, **kwargs):
        """Evaluate the metric by instance index (row).

        Parameters
        ----------
        y_true : pd.Series, pd.DataFrame, 1D np.array, or 2D np.ndarray
            Ground truth (correct) target values.

        y_pred : skpro BaseDistribution of same shape as y_true
            Predictive distribution.
            Must have same index and columns as y_true.

        C_true : pd.Series, pd.DataFrame, np.ndarray, optional (default=None)
            censoring information for survival analysis,
            should have same column name as y, same length as X and y
            should have entries 0 and 1 (float or int)
            0 = uncensored, 1 = (right) censored
            if None, all observations are assumed to be uncensored

        Returns
        -------
        loss : ``pd.Series`` or ``pd.DataFrame``
            Calculated metric, by time point (default=jackknife pseudo-values).

            ``pd.Series`` if ``self.multioutput="uniform_average"`` or array-like

            * index is equal to index of ``y_true``
            * entry at index i is metric at time i, averaged over variables

            ``pd.DataFrame`` if ``self.multioutput="raw_values"``

            * index and columns equal to those of ``y_true``
            * i,j-th entry is metric at time i, at variable j
        """
        return super().evaluate_by_index(y_true=y_true, y_pred=y_pred, **kwargs)



# LEGACY MODULE - TODO: remove or refactor

import abc

if False:
    import numpy as np
    from tabulate import tabulate

    from ..base import Controller, Model, View
    from ..cross_validation import CrossValidationController, CrossValidationView
    from ..utils import InfoController, InfoView, ItemView


class Modifier(metaclass=abc.ABCMeta):
    """Abstract modifier baseclass"""

    @abc.abstractmethod
    def modify(self, raw, headers):
        """Modifies a raw table

        Parameters
        ----------
        raw
        headers

        Returns
        -------
        Modified raw, headers
        """
        return raw, headers


class IdModifier(Modifier):
    """IdModifier

    Parameters
    ----------
    start_with: int, default=1
        Offset in ID count
    """

    def __init__(self, start_with=1):
        self.start_with = start_with

    def modify(self, raw, headers):
        for i, row in enumerate(raw):
            raw[i] = [
                {"data": {"index": self.start_with + i}, "view": ItemView("index")}
            ] + row

        headers = ["#"] + headers

        return raw, headers


class RankModifier(Modifier):
    """Rank modifier

    Parameters
    ----------
    vertical
    horizontal
    aggregate
    visible
    """

    def __init__(
        self, vertical="score", horizontal="score", aggregate=False, visible=False
    ):
        self.vertical = vertical
        self.horizontal = horizontal
        self.aggregate = aggregate
        self.visible = visible
        self.rendered_ = None

    def _rank(self, a):
        """
        Calculates a rank vector from a list

        Parameters
        ----------
        a : array-like
            The list

        Returns
        -------
        Rank vector

        Additional Information
        ----------------------
        See https://stackoverflow.com/questions/3071415/efficient-method-to-calculate-the-rank-vector-of-a-list-in-python
        """

        def rank_simple(vector):
            return sorted(range(len(vector)), key=vector.__getitem__)

        n = len(a)
        ivec = rank_simple(a)
        svec = [a[rank] for rank in ivec]
        sumranks = 0
        dupcount = 0
        result = [0] * n
        for i in range(n):
            sumranks += i
            dupcount += 1
            if i == n - 1 or svec[i] != svec[i + 1]:
                averank = sumranks / float(dupcount) + 1
                for j in range(i - dupcount + 1, i + 1):
                    result[ivec[j]] = averank
                sumranks = 0
                dupcount = 0

        return result

    def modify(self, raw, headers):
        # Calculate horizontal ranks
        if self.horizontal:
            field = self.horizontal
            for row in raw:
                sortable = []
                map = {}
                for i, cell in enumerate(row):
                    if field in cell["data"]:
                        sortable.append(cell["data"][field])
                        map[len(sortable) - 1] = cell["data"]

                for k, rank in enumerate(self._rank(sortable)):
                    map[k]["hrank"] = rank

        # Calculate vertical ranks
        if self.vertical:
            field = self.vertical
            for col in range(len(raw[0])):
                sortable = []
                map = {}
                for i in range(len(raw)):
                    cell = raw[i][col]
                    if field in cell["data"]:
                        sortable.append(cell["data"][field])
                        map[len(sortable) - 1] = cell["data"]

                for k, rank in enumerate(self._rank(sortable)):
                    map[k]["vrank"] = rank

        # Aggregate
        if self.aggregate:
            # vertical
            for row in raw:
                avg = []
                for cell in row:
                    if "vrank" in cell["data"]:
                        avg.append(cell["data"]["vrank"])

                avg_mean = np.mean(avg) if len(avg) > 0 else 0
                row.append({"data": {"vrank": avg_mean}})

            # horizontal
            avgs = []
            for col in range(len(raw[0])):
                avg = []
                for i in range(len(raw)):
                    cell = raw[i][col]
                    if "hrank" in cell["data"]:
                        avg.append(cell["data"]["hrank"])
                avg_mean = np.mean(avg) if len(avg) > 0 else 0
                avgs.append({"data": {"hrank": avg_mean}})
            raw.append(avgs)

        return raw, headers


class SortModifier(Modifier):
    """SortModifier

    Parameters
    ----------
    key
    reverse
    """

    def __init__(self, key=None, reverse=False):
        if key is None:

            def key(x):
                try:
                    return x[-1]["data"]["score"]
                except IndexError:
                    return 0
                except KeyError:
                    return 0

        self.key = key
        self.reverse = reverse

    def modify(self, raw, headers):
        sorted_raw = sorted(raw, key=self.key, reverse=self.reverse)
        return sorted_raw, headers


def filter_modifier(modifier):
    if modifier == "rank" or modifier == "ranks":
        return RankModifier()
    elif modifier == "id" or modifier == "ids":
        return IdModifier()
    elif modifier == "sort":
        return SortModifier()
    else:
        return modifier


class Table:
    """Table

    Parameters
    ----------
    tasks
    modifiers
    """

    def __init__(self, tasks=None, modifiers=None):
        if tasks is None:
            tasks = []
        self.tasks = tasks
        if modifiers is None:
            modifiers = ["id", "rank", "sort"]
        # resolve shorthands
        modifiers = [filter_modifier(modifier) for modifier in modifiers]

        self.modifiers = modifiers
        self.rendered_ = None

    def add(self, controller, view):
        """Add controllers

        Parameters
        ----------
        controller
        view

        Returns
        -------

        """
        if not issubclass(controller.__class__, Controller):
            raise Exception(
                "controller has to be subclass instance skpro.workflow.Controller"
            )

        if not issubclass(view.__class__, View):
            raise Exception("view has to be subclass instance of skpro.workflow.View")

        self.tasks.append((controller, view))
        self.rendered_ = None

        return self

    def cv(
        self,
        data,
        loss_func,
        tune=False,
        cv=None,
        optimizer=None,
        display_tuning=False,
        with_ranks=True,
    ):
        """

        Parameters
        ----------
        data
        loss_func
        tune
        cv
        with_tuning
        with_ranks

        Returns
        -------

        """
        return self.add(
            CrossValidationController(
                data, loss_func, cv=cv, tune=tune, optimizer=optimizer
            ),
            CrossValidationView(with_tuning=display_tuning, with_ranks=with_ranks),
        )

    def info(self, with_group=False):
        """

        Parameters
        ----------
        with_group

        Returns
        -------

        """
        return self.add(InfoController(), InfoView(with_group=with_group))

    def modify(self, modifier):
        self.modifiers.append(filter_modifier(modifier))

    def render(self, models, verbose=0, debug=False):
        """Render table

        Parameters
        ----------
        models
        verbose
        debug

        Returns
        -------

        """

        if len(self.tasks) == 0:
            raise Exception("The table is empty. You have to include tasks using add()")

        # Compose table
        raw_tbl = []
        headers = []
        i = 0
        for model in models:
            if not issubclass(model.__class__, Model):
                model = Model(model)

            if verbose > 1:
                print("Evaluating model %i/%i: %s" % (i + 1, len(models), str(model)))
            row = []
            for controller, view in self.tasks:
                controller_id = controller.identifier()
                if not debug or str(controller) == "Info":
                    model[controller_id] = controller.run(model)
                else:
                    model[controller_id] = {"debug": "debug"}
                    view = ItemView("debug")

                if verbose > 2:
                    print("Running " + str(controller))
                    print(model[controller_id])

                if i == 0:
                    headers.append(str(controller))

                row.append({"data": model[controller_id], "view": view})
            i += 1
            raw_tbl.append(row)

        # Apply modifiers
        for modifier in self.modifiers:
            raw_tbl, headers = modifier.modify(raw_tbl, headers)

        # Parse raw table
        tbl = []
        for raw in raw_tbl:
            tbl.append(
                [cell["view"].parse(cell["data"]) for cell in raw if "view" in cell]
            )

        self.rendered_ = {"raw": raw_tbl, "parsed": tbl, "headers": headers}

        return self.rendered_

    def print(
        self,
        models,
        fmt="pipe",
        with_headers=True,
        raw=False,
        return_only=False,
        verbose=1,
        debug=False,
    ):
        """Print table

        Parameters
        ----------
        models
        fmt: string or list
            None, 'pipe', 'latex'
        with_headers
        raw
        return_only
        verbose
        debug

        Returns
        -------

        """

        self.render(models, verbose=verbose, debug=debug)

        if with_headers:
            headers = self.rendered_["headers"]
        else:
            headers = []

        if raw:
            rendered = self.rendered_["raw"]
        else:
            rendered = self.rendered_["parsed"]

        if isinstance(fmt, str):
            fmt = [fmt]

        tbl = ""
        for f in fmt:
            if f == "raw":
                f = "pipe"
                rendered = self.rendered_["raw"]

            tbl += (
                "\n"
                + tabulate(
                    rendered,
                    headers=headers,
                    tablefmt=f,
                )
                + "\n"
            )

        if not return_only:
            print(tbl)

        return tbl



"""Test class for default methods.

This is not for direct use, but for testing whether the defaulting in various
methods works.

Testing works via TestAllDistributions which discovers the classes in
here, executes the public methods in interface conformance tests,
which in turn triggers the fallback defaults.
"""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
# adapted from sktime

__author__ = ["fkiraly"]

import numpy as np
import pandas as pd
import pytest
from scipy.special import erfinv

from skpro.distributions.base import BaseDistribution
from skpro.tests.test_switch import run_test_module_changed
from skpro.utils.estimator_checks import check_estimator


# normal distribution with exact implementations removed
class _DistrDefaultMethodTester(BaseDistribution):
    """Tester distribution for default methods."""

    _tags = {
        "capabilities:approx": ["pdfnorm", "mean", "var", "energy", "log_pdf", "cdf"],
        "capabilities:exact": ["pdf", "ppf"],
        "distr:measuretype": "continuous",
        "broadcast_init": "on",
    }

    def __init__(self, mu, sigma, index=None, columns=None):
        self.mu = mu
        self.sigma = sigma

        super().__init__(index=index, columns=columns)

    def _ppf(self, p):
        """Quantile function = percent point function = inverse cdf.

        Parameters
        ----------
        p : 2D np.ndarray, same shape as ``self``
            values to evaluate the ppf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            ppf values at the given points
        """
        mu = self._bc_params["mu"]
        sigma = self._bc_params["sigma"]

        icdf_arr = mu + sigma * np.sqrt(2) * erfinv(2 * p - 1)
        return icdf_arr

    def _pdf(self, x):
        """Probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            pdf values at the given points
        """
        mu = self._bc_params["mu"]
        sigma = self._bc_params["sigma"]

        pdf_arr = np.exp(-0.5 * ((x - mu) / sigma) ** 2)
        pdf_arr = pdf_arr / (sigma * np.sqrt(2 * np.pi))
        return pdf_arr

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {"mu": [[0, 1], [2, 3], [4, 5]], "sigma": 1}
        params2 = {
            "mu": 0,
            "sigma": 1,
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),
        }
        # scalar case examples
        params3 = {"mu": 1, "sigma": 2}
        return [params1, params2, params3]


@pytest.mark.skipif(
    not run_test_module_changed("skpro.distributions"),
    reason="run only if skpro.distributions has been changed",
)
def test_base_default():
    """Test default methods.

    The _DistributionDefaultMethodTester class is not detected
    by TestAllDistributions (it is private), so we need to test it explicitly.

    check_estimator invokes a TestAllDistributions call.
    """
    check_estimator(_DistrDefaultMethodTester, raise_exceptions=True)



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Chi-Squared probability distribution."""

__author__ = ["sukjingitsit"]

import pandas as pd
from scipy.stats.distributions import chi2

from skpro.distributions.base import BaseDistribution


class ChiSquared(BaseDistribution):
    """Chi-Squared distribution (skpro native).

    Parameters
    ----------
    dof : float or array of float (1D or 2D)
        degrees of freedom of the chi-squared distribution
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions.chi_squared import ChiSquared
    >>> chi = ChiSquared(dof=[[1, 2], [3, 4], [5, 6]])
    """

    _tags = {
        # packaging info
        # --------------
        "authors": "sukjingitsit",
        # estimator tags
        # --------------
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf", "ppf"],
        "distr:measuretype": "continuous",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def __init__(self, dof, index=None, columns=None):
        self.dof = dof

        super().__init__(index=index, columns=columns)

    r"""Energy implementation issues:

    The self-energy is mathematically difficult to calculate due to
    their being no proper closed form. As discussed with fkiraly,
    using E(d.energy(x)) is one possible way, but the question arises
    on how to approximate the integral. The other alternative is to use
    sampling to estimate the self-energy.

    The closed form version for cross-energy can be framed as follows:
    Here, :math:`k=dof`
    :math:`x <= 0, \operatorname{energy}(x) = k + \vert x \vert`
    :math:`x > 0, \operatorname{energy}(x) =
    x*(2*\operatorname{CDF}(k,x)-1)+k-2k*\operatorname{CDF}(k+1,x)`
    where :math:`\operatorname{CDF}(k,x)` represents the CDF of x
    for a chi-square distribution with k degrees of freedom.
    """

    def _mean(self):
        """Return expected value of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            expected value of distribution (entry-wise)
        """
        return self._bc_params["dof"]

    def _var(self):
        r"""Return element/entry-wise variance of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            variance of the distribution (entry-wise)
        """
        return 2 * self._bc_params["dof"]

    def _pdf(self, x):
        """Probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            pdf values at the given points
        """
        dof = self._bc_params["dof"]
        pdf_arr = chi2.pdf(x, dof)
        return pdf_arr

    def _log_pdf(self, x):
        """Logarithmic probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            log pdf values at the given points
        """
        dof = self._bc_params["dof"]
        lpdf_arr = chi2.logpdf(x, dof)
        return lpdf_arr

    def _cdf(self, x):
        """Cumulative distribution function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the cdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            cdf values at the given points
        """
        dof = self._bc_params["dof"]
        cdf_arr = chi2.cdf(x, dof)
        return cdf_arr

    def _ppf(self, p):
        """Quantile function = percent point function = inverse cdf.

        Parameters
        ----------
        p : 2D np.ndarray, same shape as ``self``
            values to evaluate the ppf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            ppf values at the given points
        """
        dof = self._bc_params["dof"]
        icdf_arr = chi2.ppf(p, dof)
        return icdf_arr

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {"dof": [[1, 2], [3, 4], [5, 6]]}
        params2 = {
            "dof": 10,
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),
        }
        # scalar case examples
        params3 = {"dof": 3}
        return [params1, params2, params3]



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Poisson probability distribution."""

__author__ = ["fkiraly", "malikrafsan"]

import pandas as pd
from scipy.stats import poisson, rv_discrete

from skpro.distributions.adapters.scipy import _ScipyAdapter


class Poisson(_ScipyAdapter):
    """Poisson distribution.

    Most methods wrap ``scipy.stats.poisson``.

    Parameters
    ----------
    mu : float or array of float (1D or 2D)
        mean of the distribution
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions import Poisson

    >>> distr = Poisson(mu=[[1, 1], [2, 3], [4, 5]])
    """

    _tags = {
        "capabilities:approx": ["ppf", "energy"],
        "capabilities:exact": ["mean", "var", "pmf", "log_pmf", "cdf"],
        "distr:measuretype": "discrete",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def __init__(self, mu, index=None, columns=None):
        self.mu = mu

        super().__init__(index=index, columns=columns)

    def _get_scipy_object(self) -> rv_discrete:
        return poisson

    def _get_scipy_param(self) -> dict:
        mu = self._bc_params["mu"]

        return [mu], {}

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        params1 = {"mu": [[1, 1], [2, 3], [4, 5]]}
        params2 = {
            "mu": 0,
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),
        }
        return [params1, params2]



"""Interface adapters to scikit-survival tree model."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["fkiraly"]

from skpro.survival.adapters.sksurv import _SksurvAdapter
from skpro.survival.base import BaseSurvReg


class SurvivalTree(_SksurvAdapter, BaseSurvReg):
    """Survival tree, from scikit-survival.

    Direct interface to ``sksurv.tree.SurvivalTree``, by ``sebp``.

    The quality of a split is measured by the log-rank splitting rule.

    If ``splitter='best'``, fit and predict methods support missing values.

    See [1]_, [2]_ and [3]_ for further description.

    Parameters
    ----------
    splitter : {'best', 'random'}, default: 'best'
        The strategy used to choose the split at each node. Supported
        strategies are 'best' to choose the best split and 'random' to choose
        the best random split.

    max_depth : int or None, optional, default: None
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        `min_samples_split` samples.

    min_samples_split : int, float, optional, default: 6
        The minimum number of samples required to split an internal node:

        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.

    min_samples_leaf : int, float, optional, default: 3
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.

        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.

    min_weight_fraction_leaf : float, optional, default: 0.
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when sample_weight is not provided.

    max_features : int, float, string or None, optional, default: None
        The number of features to consider when looking for the best split:

        - If int, then consider `max_features` features at each split.
        - If float, then `max_features` is a fraction and
          `max(1, int(max_features * n_features_in_))` features are considered at
          each split.
        - If "sqrt", then `max_features=sqrt(n_features)`.
        - If "log2", then `max_features=log2(n_features)`.
        - If None, then `max_features=n_features`.

        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.

    random_state : int, RandomState instance or None, optional, default: None
        Controls the randomness of the estimator. The features are always
        randomly permuted at each split, even if ``splitter`` is set to
        ``"best"``. When ``max_features < n_features``, the algorithm will
        select ``max_features`` at random at each split before finding the best
        split among them. But the best found split may vary across different
        runs, even if ``max_features=n_features``. That is the case, if the
        improvement of the criterion is identical for several splits and one
        split has to be selected at random. To obtain a deterministic behavior
        during fitting, ``random_state`` has to be fixed to an integer.

    max_leaf_nodes : int or None, optional, default: None
        Grow a tree with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    Attributes
    ----------
    unique_times_ : array of shape = (n_unique_times,)
        Unique time points.

    max_features_ : int,
        The inferred value of max_features.

    tree_ : Tree object
        The underlying Tree object. Please refer to
        ``help(sklearn.tree._tree.Tree)`` for attributes of Tree object.

    References
    ----------
    .. [1] Leblanc, M., & Crowley, J. (1993). Survival Trees by Goodness of Split.
           Journal of the American Statistical Association, 88(422), 457–467.

    .. [2] Ishwaran, H., Kogalur, U. B., Blackstone, E. H., & Lauer, M. S. (2008).
           Random survival forests. The Annals of Applied Statistics, 2(3), 841–860.

    .. [3] Ishwaran, H., Kogalur, U. B. (2007). Random survival forests for R.
           R News, 7(2), 25–31. https://cran.r-project.org/doc/Rnews/Rnews_2007-2.pdf.
    """

    _tags = {
        "authors": ["sebp", "fkiraly"],  # sebp credit for interfaced estimator
        "capability:missing": True,  # only for splitter="best"
    }

    def __init__(
        self,
        splitter="best",
        max_depth=None,
        min_samples_split=6,
        min_samples_leaf=3,
        min_weight_fraction_leaf=0.0,
        max_features=None,
        random_state=None,
        max_leaf_nodes=None,
    ):
        self.splitter = splitter
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.random_state = random_state
        self.max_leaf_nodes = max_leaf_nodes

        super().__init__()

        if splitter != "best":
            self.set_tags(**{"capability:missing": False})

    def _get_sksurv_class(self):
        """Getter of the sksurv class to be used for the adapter."""
        from sksurv.tree import SurvivalTree as _SurvivalTree

        return _SurvivalTree

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params1 = {}

        params2 = {
            "splitter": "random",
            "max_depth": 3,
            "min_samples_split": 3,
            "min_samples_leaf": 2,
        }

        return [params1, params2]



"""Adapters for probability distribution objects, statsmodels facing."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

from skpro.distributions.adapters.statsmodels._empirical import empirical_from_rvdf

__all__ = ["empirical_from_rvdf"]



"""Probability distribution objects."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
# adapted from sktime

__all__ = ["BaseDistribution", "_DelegatedDistribution", "_BaseArrayDistribution"]

from skpro.distributions.base._base import BaseDistribution
from skpro.distributions.base._base_array import _BaseArrayDistribution
from skpro.distributions.base._delegate import _DelegatedDistribution



"""Reduction to probabilistic classification."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

from skpro.regression.binned._sklearn_bin_regressor import HistBinnedProbaRegressor

__all__ = ["HistBinnedProbaRegressor"]



# -*- coding: utf-8 -*-

import numpy as np
from sklearn.base import BaseEstimator

from skpro.utils.utils import to_percent


class Minimum(BaseEstimator):
    """Minimum estimator

    Wrapping estimator that replaces predictions of the wrapped
    estimator that fall below a specified minimum threshold
    with the threshold itself.

    Parameters
    ----------
    estimator: subclass of sklearn.base.BaseEstimator
        Estimator which predicts shall be bounded by minimum threshold
    minimum: float/int
        Minimum boundary for the estimator's predictions. If relative=True
        the minimum represent a percentage value
    relative: bool
        If true, minimum will be regarded as percentage value
        and the cut-off threshold will be determined dynamically
        during fitting as ``threshold = minimum * std(y)``

    Properties
    ----------
    estimator : subclass of sklearn.base.BaseEstimator
        Wrapped estimator
    minimum : float
        Minimum threshold
    relative: bool
        If minimum is relative with regard to label variance
    """

    def __init__(self, base_estimator, minimum=0.3, relative=True):
        self.base_estimator = base_estimator
        if relative:
            self.minimum = to_percent(minimum)
        else:
            self.minimum = minimum
        self.relative = relative

    def fit(self, X, y):
        # Forward fitting to wrapped estimator
        if getattr(self, "estimator", False):
            self.base_estimator.estimator = self.estimator
        self.base_estimator.fit(X, y)

        # Compute absolute minimum
        if self.relative:
            self.minimum *= np.std(y)

        return self

    def predict(self, X):
        # Apply cut-off
        prediction = self.base_estimator.predict(X)
        prediction[prediction < self.minimum] = self.minimum

        return prediction

    def __str__(self, describer=str):
        return (
            "Min(" + describer(self.base_estimator) + ", min=" + str(self.minimum) + ")"
        )

    def __repr__(self):
        return self.__str__(repr)


class Constant(BaseEstimator):
    """Constant estimator

    Predicts predefinied constant

    Parameters
    ----------
    constant: float | callable(X, y) | string: 'mean(y)', 'std(y)' (default: None)
        Specifies the constant. A callable receives the training data during
        fit and should return a constant value. The string options provide
        a shortcut for mean/std extraction from the features.
    name: string (optional)
        Optional description of the constant for the estimator string
        representation. Defaults to str(constant).
    """

    def __init__(self, constant=None, name=None):
        self.constant = constant
        self.name = name

    def fit(self, X, y):
        # evaluate callables
        if callable(self.constant):
            self.constant = self.constant(X, y)

        # resolve str shorthands
        if isinstance(self.constant, str):
            self.name = self.constant
            if self.constant == "mean(y)":
                self.constant = np.mean(y)
            elif self.constant == "std(y)":
                self.constant = np.std(y)
            else:
                raise ValueError(self.constant + " is not a valid function")

        # if no constant was provided we use the y value
        if self.constant is None:
            self.constant = y

        return self

    def predict(self, X):
        return np.ones((X.shape[0],)) * self.constant

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        if self.name is None:
            return "C(" + str(self.constant) + ")"
        else:
            return "C(" + self.name + ")"



"""Tests for adapters for probability distribution objects, scipy facing."""

import numpy as np
import pandas as pd
import pytest
from skbase.testing import QuickTester

from skpro.distributions.adapters.scipy import _ScipyAdapter
from skpro.tests.test_all_estimators import BaseFixtureGenerator, PackageConfig

__author__ = ["fkiraly", "malikrafsan"]


def test_empirical_from_discrete():
    """Test empirical_from_discrete."""
    from scipy.stats import rv_discrete

    from skpro.distributions.adapters.scipy._empirical import empirical_from_discrete

    xk = np.arange(7)
    pk = (0.1, 0.2, 0.3, 0.1, 0.1, 0.0, 0.2)
    pk2 = (0.1, 0.1, 0.4, 0.0, 0.1, 0.2, 0.1)

    dist1 = rv_discrete(name="custm", values=(xk, pk))
    dist2 = rv_discrete(name="custm", values=(xk, pk2))

    emp = empirical_from_discrete([dist1, dist2])
    assert isinstance(emp.spl, pd.DataFrame)
    assert isinstance(emp.weights, pd.Series)
    assert emp.spl.shape == (14, 1)
    assert emp.weights.shape == (14,)
    expected_idx = pd.MultiIndex.from_arrays(
        [[0, 1, 2, 3, 4, 5, 6] * 2, [0] * 7 + [1] * 7]
    )
    assert np.all(emp.spl.index == expected_idx)
    assert np.all(emp.spl.columns == [0])

    emp2 = empirical_from_discrete(
        [dist1, dist2], index=pd.Index(["foo", "bar"]), columns=["abc"]
    )
    assert isinstance(emp2.spl, pd.DataFrame)
    assert isinstance(emp2.weights, pd.Series)
    assert emp2.spl.shape == (14, 1)
    assert emp2.weights.shape == (14,)
    expected_idx = pd.MultiIndex.from_arrays(
        [[0, 1, 2, 3, 4, 5, 6] * 2, ["foo"] * 7 + ["bar"] * 7]
    )
    assert np.all(emp2.spl.index == expected_idx)
    assert np.all(emp2.spl.columns == ["abc"])


class ScipyDistributionFixtureGenerator(BaseFixtureGenerator):
    """Fixture generator for scipy distributions adapter.

    Fixtures parameterized
    ----------------------
    object_class: object inheriting from BaseObject
        ranges over object classes not excluded by EXCLUDE_OBJECTS, EXCLUDED_TESTS
    object_instance: instance of object inheriting from BaseObject
        ranges over object classes not excluded by EXCLUDE_OBJECTS, EXCLUDED_TESTS
        instances are generated by create_test_instance class method
    """

    object_type_filter = _ScipyAdapter


class TestScipyAdapter(PackageConfig, ScipyDistributionFixtureGenerator, QuickTester):
    """Test the scipy adapter."""

    METHOD_TESTS = {
        "NO_PARAMS": [("mean", "mean"), ("var", "var")],
        "X_PARAMS": [("cdf", "cdf"), ("ppf", "ppf")],
        "CONTINUOUS": [("pdf", "pdf"), ("log_pdf", "logpdf")],
        "DISCRETE": [("pmf", "pmf"), ("log_pmf", "logpmf")],
    }

    X_VALUES = [0.1, 0.5, 0.99]

    @pytest.mark.parametrize("method,scipy_method", METHOD_TESTS["NO_PARAMS"])
    def test_method_no_params(self, object_instance, method, scipy_method):
        """Test method that doesn't need additional parameters."""
        res = getattr(object_instance, method)()
        params = object_instance._get_scipy_param()
        scipy_obj = object_instance._get_scipy_object()

        scipy_res = getattr(scipy_obj, scipy_method)(*params[0], **params[1])

        assert np.allclose(res, scipy_res)

    @pytest.mark.parametrize("method,scipy_method", METHOD_TESTS["X_PARAMS"])
    @pytest.mark.parametrize("x", X_VALUES)
    def test_method_with_x_params(self, object_instance, method, scipy_method, x):
        """Test method that needs x as parameter."""
        res = getattr(object_instance, method)(x)
        params = object_instance._get_scipy_param()
        scipy_obj = object_instance._get_scipy_object()

        scipy_res = getattr(scipy_obj, scipy_method)(x, *params[0], **params[1])

        assert np.allclose(res, scipy_res)

    @pytest.mark.parametrize("method,scipy_method", METHOD_TESTS["CONTINUOUS"])
    @pytest.mark.parametrize("x", X_VALUES)
    def test_method_continuous_dist(self, object_instance, method, scipy_method, x):
        """Test continuous distribution method."""
        res = getattr(object_instance, method)(x)
        if object_instance._tags["distr:measuretype"] != "continuous":
            return None  # in this case, scipy method is not defined
        else:
            params = object_instance._get_scipy_param()
            scipy_obj = object_instance._get_scipy_object()
            scipy_res = getattr(scipy_obj, scipy_method)(x, *params[0], **params[1])

        assert np.allclose(res, scipy_res)

    @pytest.mark.parametrize("method,scipy_method", METHOD_TESTS["DISCRETE"])
    @pytest.mark.parametrize("x", X_VALUES)
    def test_method_discrete_dist(self, object_instance, method, scipy_method, x):
        """Test discrete distribution method."""
        res = getattr(object_instance, method)(x)
        if object_instance._tags["distr:measuretype"] != "discrete":
            return None  # in this case, scipy method is not defined
        else:
            params = object_instance._get_scipy_param()
            scipy_obj = object_instance._get_scipy_object()
            scipy_res = getattr(scipy_obj, scipy_method)(x, *params[0], **params[1])

        assert np.allclose(res, scipy_res)



"""Linear regression models."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

from skpro.regression.dummy import DummyProbaRegressor
from skpro.regression.linear._glm import GLMRegressor
from skpro.regression.linear._sklearn import ARDRegression, BayesianRidge
from skpro.regression.linear._sklearn_poisson import PoissonRegressor

__all__ = [
    "ARDRegression",
    "BayesianRidge",
    "GLMRegressor",
    "PoissonRegressor",
    "DummyProbaRegressor",
]



"""Tests for cyclic boosting regressor."""

import pandas as pd
import pytest

from skpro.regression.cyclic_boosting import CyclicBoosting
from skpro.tests.test_switch import run_test_for_class


@pytest.mark.skipif(
    not run_test_for_class(CyclicBoosting),
    reason="run test only if softdeps are present and incrementally (if requested)",
)
def test_cyclic_boosting_simple_use():
    """Test simple use of cyclic boosting regressor."""
    from sklearn.datasets import load_diabetes
    from sklearn.model_selection import train_test_split

    X, y = load_diabetes(return_X_y=True, as_frame=True)
    y = pd.DataFrame(y)
    X = X.iloc[:200]
    y = y.iloc[:200]
    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=0)

    reg_proba = CyclicBoosting()
    reg_proba.fit(X_train, y_train)
    y_pred = reg_proba.predict_proba(X_test)

    assert y_pred.shape == y_test.shape


@pytest.mark.skipif(
    not run_test_for_class(CyclicBoosting),
    reason="run test only if softdeps are present and incrementally (if requested)",
)
def test_cyclic_boosting_with_manual_parameters():
    """Test use of cyclic boosting regressor with_manual_parameters."""
    from cyclic_boosting import flags
    from sklearn.datasets import load_diabetes
    from sklearn.model_selection import train_test_split

    X, y = load_diabetes(return_X_y=True, as_frame=True)
    y = pd.DataFrame(y)
    X = X.iloc[:200]
    y = y.iloc[:200]
    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=0)

    features = [
        "age",
        "sex",
        "bmi",
        "bp",
        "s1",
        "s2",
        "s3",
        ("age", "sex"),
    ]

    fp = {
        "age": flags.IS_UNORDERED,
        "sex": flags.IS_UNORDERED,
        "bmi": flags.IS_CONTINUOUS,
        "bp": flags.IS_CONTINUOUS,
        "s1": flags.IS_CONTINUOUS,
        "s2": flags.IS_CONTINUOUS,
        "s3": flags.IS_CONTINUOUS,
    }

    reg_proba = CyclicBoosting(
        feature_groups=features,
        feature_properties=fp,
        maximal_iterations=5,
        alpha=0.25,
        mode="additive",
        lower=0.0,
    )
    reg_proba.fit(X_train, y_train)
    y_pred = reg_proba.predict_proba(X_test)

    assert y_pred.shape == y_test.shape



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Histogram distribution."""

__author__ = ["ShreeshaM07"]

import numpy as np
import pandas as pd

from skpro.distributions.base import _BaseArrayDistribution


class Histogram(_BaseArrayDistribution):
    """Histogram Probability Distribution.

    The histogram probability distribution is parameterized
    by the ``bins`` and ``bin_mass``.

    Parameters
    ----------
    bins : tuple(float, float, int) or numpy.array of float 1D or 2D list of size m x n
        1. tuple(first bin's start point, last bin's end point, number of bins)
           Used when ``bin widths`` are ``equal``.
            example:
                    bins:(0,4,4)
        2. array has the bin boundaries with 1st element the first bin's
           starting point and rest are the bin ending points of all bins
            example:
                    bins:[0, 1, 2, 3, 4]
        3. 2D list of size m x n containing m*n float numpy.arrays or tuple
           like ``case 1``.
            example:
                bins:
                [
                        [[0, 1, 2, 3, 4], [5, 5.5, 5.8, 6.5, 7, 7.5]],
                        [(2, 12, 5), [0, 1, 2, 3, 4]],
                        [[1.5, 2.5, 3.1, 4, 5.4], [-4, -2, -1.5, 5, 10]]
                ]
    bin_mass : array of float 1D or 2D list of size m x n
        1. Array has the mass of the bins or area of the bins.
           example:
                bin_mass:[0.1, 0.2, 0.3, 0.4]
           ``Note``: ``len(bin_mass)`` will be ``(len(bins)-1)``.
           ``Note``: ``sum(bin_mass)`` must be ``1``.
        2. 2D list of size m x n containing m*n float numpy.arrays
           each satisfying ``case 1``.
            example:
                bin_mass:
                [
                        [[0.1, 0.2, 0.3, 0.4], [0.25, 0.1, 0, 0.4, 0.25]],
                        [[0.1, 0.2, 0.4, 0.2, 0.1], [0.4, 0.3, 0.2, 0.1]],
                        [[0.06, 0.15, 0.09, 0.7], [0.4, 0.15, 0.325, 0.125]]
                ]
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex
    """

    _tags = {
        "authors": ["ShreeshaM07"],
        "capabilities:approx": ["pdfnorm", "energy"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf", "ppf"],
        "distr:measuretype": "continuous",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def _convert_tuple_to_array(self, bins):
        bins_to_list = (bins[0], bins[1], bins[2])
        bins = []
        bin_width = (bins_to_list[1] - bins_to_list[0]) / bins_to_list[2]
        for b in range(bins_to_list[2]):
            bins.append(bins_to_list[0] + b * bin_width)
        bins.append(bins_to_list[1])
        return bins

    def _check_single_array_distr(self, bins, bin_mass):
        all1Ds = (
            isinstance(bins[0], float)
            or isinstance(bins[0], np.floating)
            or isinstance(bins[0], int)
            or isinstance(bins[0], np.integer)
        )
        all1Ds = (
            all1Ds
            and isinstance(bin_mass[0], int)
            or isinstance(bin_mass[0], np.integer)
            or isinstance(bin_mass[0], float)
            or isinstance(bin_mass[0], np.floating)
            and np.array(bin_mass).ndim == 1
        )
        return all1Ds

    def __init__(self, bins, bin_mass, index=None, columns=None):
        if isinstance(bins, tuple):
            bins = self._convert_tuple_to_array(bins)
            self.bins = np.array(bins)
            self.bin_mass = np.array(bin_mass)
        elif self._check_single_array_distr(bins, bin_mass):
            self.bins = np.array(bins)
            self.bin_mass = np.array(bin_mass)
        else:
            # convert the bins into a list
            for i in range(len(bins)):
                for j in range(len(bins[i])):
                    if isinstance(bins[i][j], tuple):
                        bins[i][j] = self._convert_tuple_to_array(bins[i][j])
                    bins[i][j] = np.array(bins[i][j])
                    bin_mass[i][j] = np.array(bin_mass[i][j])
            self.bins = bins
            self.bin_mass = bin_mass

        super().__init__(index=index, columns=columns)

    def _energy_self(self):
        r"""Energy of self, w.r.t. self.

        :math:`\mathbb{E}[|X-Y|]`, where :math:`X, Y` are i.i.d. copies of self.

        Private method, to be implemented by subclasses.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            energy values w.r.t. the given points
        """
        bins = self.bins
        bin_mass = self.bin_mass
        energy_arr = []
        from numpy.lib.stride_tricks import sliding_window_view

        if self._check_single_array_distr(bins, bin_mass):
            bins_hist = bins
            bin_mass_hist = bin_mass
            win_centre_bins = 0.5 * np.sum(
                sliding_window_view(bins_hist, window_shape=2), axis=1
            )
            expected_value = 0
            for i in range(len(bin_mass_hist)):
                for j in range(len(bin_mass_hist)):
                    expected_value += (
                        bin_mass_hist[i]
                        * bin_mass_hist[j]
                        * abs(win_centre_bins[i] - win_centre_bins[j])
                    )
            energy_arr = expected_value
            return expected_value

        for row in range(len(bins)):
            energy_arr_row = []
            for col in range(len(bins[0])):
                bins_hist = bins[row][col]
                bin_mass_hist = bin_mass[row][col]
                win_centre_bins = 0.5 * np.sum(
                    sliding_window_view(bins_hist, window_shape=2), axis=1
                )
                expected_value = 0
                for i in range(len(bin_mass_hist)):
                    for j in range(len(bin_mass_hist)):
                        expected_value += (
                            bin_mass_hist[i]
                            * bin_mass_hist[j]
                            * abs(win_centre_bins[i] - win_centre_bins[j])
                        )
                energy_arr_row.append(expected_value)
            energy_arr.append(energy_arr_row)
        energy_arr = np.array(energy_arr)
        if energy_arr.ndim > 0:
            energy_arr = np.sum(energy_arr, axis=1)
        return energy_arr

    def _energy_x(self, x):
        r"""Energy of self, w.r.t. a constant frame x.

        :math:`\mathbb{E}[|X-x|]`, where :math:`X` is a copy of self,
        and :math:`x` is a constant.

        Private method, to be implemented by subclasses.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to compute energy w.r.t. to

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            energy values w.r.t. the given points
        """
        bins = self.bins
        bin_mass = self.bin_mass
        energy_arr = []
        mean = self._mean()
        cdf = self._cdf(x)
        pdf = self._pdf(x)
        x = np.array(x)
        from numpy.lib.stride_tricks import sliding_window_view

        if self._check_single_array_distr(bins, bin_mass):
            bins_hist = np.array(bins)
            bin_mass_hist = np.array(bin_mass)
            X = x
            is_outside = X < bins_hist[0] or X > bins_hist[-1]
            if is_outside:
                energy_arr = abs(mean - X)
                return energy_arr
            else:
                # consider X lies in kth bin
                # so kth bin's start index is
                k_1_bins = np.where(X >= bins_hist)[0][-1]
                win_sum_bins = np.sum(
                    sliding_window_view(bins_hist, window_shape=2), axis=1
                )
                # upto kth bin excluding kth
                X_upto_k = X * cdf - 0.5 * np.dot(
                    win_sum_bins[:k_1_bins], bin_mass_hist[:k_1_bins]
                )
                # if X is in last bin
                if k_1_bins >= len(bin_mass_hist) - 1:
                    energy_arr = X_upto_k
                    return energy_arr
                # in the kth bin
                X_in_k = (
                    0.5
                    * pdf
                    * (
                        bins_hist[k_1_bins] ** 2
                        + bins_hist[k_1_bins + 1] ** 2
                        - 2 * X**2
                    )
                )
                # after kth bin excluding kth
                X_after_k = 0.5 * np.dot(
                    win_sum_bins[k_1_bins + 1 :], bin_mass_hist[k_1_bins + 1 :]
                ) - X * (1 - cdf)
                energy_arr = X_upto_k + X_in_k + X_after_k
                return energy_arr

        for i in range(len(bins)):
            energy_arr_row = []
            for j in range(len(bins[0])):
                bins_hist = bins[i][j]
                bin_mass_hist = bin_mass[i][j]
                X = x[i][j]
                is_outside = X < bins_hist[0] or X > bins_hist[-1]
                if is_outside:
                    energy_arr_row.append(abs(mean[i][j] - X))
                else:
                    # consider X lies in kth bin
                    # so kth bin's start index is
                    k_1_bins = np.where(X >= bins_hist)[0][-1]
                    win_sum_bins = np.sum(
                        sliding_window_view(bins_hist, window_shape=2), axis=1
                    )
                    # upto kth bin excluding kth
                    X_upto_k = X * cdf[i][j] - 0.5 * np.dot(
                        win_sum_bins[:k_1_bins], bin_mass_hist[:k_1_bins]
                    )
                    # if X is in last bin
                    if k_1_bins >= len(bin_mass_hist) - 1:
                        energy_arr_row.append(X_upto_k)
                        continue
                    # in the kth bin
                    X_in_k = (
                        0.5
                        * pdf[i][j]
                        * (
                            bins_hist[k_1_bins] ** 2
                            + bins_hist[k_1_bins + 1] ** 2
                            - 2 * X**2
                        )
                    )
                    # after kth bin excluding kth
                    X_after_k = 0.5 * np.dot(
                        win_sum_bins[k_1_bins + 1 :], bin_mass_hist[k_1_bins + 1 :]
                    ) - X * (1 - cdf[i][j])
                    energy_arr_row.append(X_upto_k + X_in_k + X_after_k)
            energy_arr.append(energy_arr_row)
        energy_arr = np.array(energy_arr)
        if energy_arr.ndim > 0:
            energy_arr = np.sum(energy_arr, axis=1)
        return energy_arr

    def _mean(self):
        """Return expected value of the distribution.

        Returns
        -------
        float, sum(bin_mass)/range(bins)
            expected value of distribution (entry-wise)
        """
        bins = self.bins
        bin_mass = self.bin_mass
        mean = []
        from numpy.lib.stride_tricks import sliding_window_view

        if self._check_single_array_distr(bins, bin_mass):
            bins = np.array(bins)
            bin_mass = np.array(bin_mass)
            win_sum_bins = np.sum(sliding_window_view(bins, window_shape=2), axis=1)
            mean = 0.5 * np.dot(win_sum_bins, bin_mass)
            return mean

        for i in range(len(bins)):
            mean_row = []
            for j in range(len(bins[0])):
                bins_hist = bins[i][j]
                bin_mass_hist = bin_mass[i][j]
                win_sum_bins = np.sum(
                    sliding_window_view(bins_hist, window_shape=2), axis=1
                )
                mean_row.append(0.5 * np.dot(win_sum_bins, bin_mass_hist))
            mean.append(mean_row)
        return np.array(mean)

    def _var(self):
        r"""Return element/entry-wise variance of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            variance of the distribution (entry-wise)
        """
        bins = self.bins
        bin_mass = self.bin_mass
        var = []
        mean = self._mean()
        from numpy.lib.stride_tricks import sliding_window_view

        if self._check_single_array_distr(bins, bin_mass):
            bins = np.array(bins)
            bin_mass = np.array(bin_mass)
            win_sum_bins = np.sum(sliding_window_view(bins, window_shape=2), axis=1)
            win_prod_bins = np.prod(sliding_window_view(bins, window_shape=2), axis=1)
            var = np.dot(bin_mass / 3, (win_sum_bins**2 - win_prod_bins)) - mean**2
            return var

        for i in range(len(bins)):
            var_row = []
            for j in range(len(bins[0])):
                bins_hist = bins[i][j]
                bin_mass_hist = bin_mass[i][j]
                win_sum_bins = np.sum(
                    sliding_window_view(bins_hist, window_shape=2), axis=1
                )
                win_prod_bins = np.prod(
                    sliding_window_view(bins_hist, window_shape=2), axis=1
                )
                var_row.append(
                    np.dot(bin_mass_hist / 3, (win_sum_bins**2 - win_prod_bins))
                    - mean[i][j] ** 2
                )
            var.append(var_row)
        var = np.array(var)
        return var

    def _pdf(self, x):
        """Probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            pdf values at the given points
        """
        bin_mass = self.bin_mass
        bins = self.bins
        pdf = []
        if self._check_single_array_distr(bins, bin_mass):
            bins = np.array(bins)
            bin_mass = np.array(bin_mass)
            bin_width = np.diff(bins)
            pdf_arr = bin_mass / bin_width
            X = x
            if len(np.where(X < bins)[0]) and len(np.where(X >= bins)[0]):
                pdf = pdf_arr[min(np.where(X < bins)[0]) - 1]
            else:
                pdf = 0
            return pdf
        # bins_hist contains the bins edges of each histogram
        for i in range(len(bins)):
            pdf_row = []
            for j in range(len(bins[i])):
                bins_hist = bins[i][j]
                bin_mass_hist = bin_mass[i][j]
                bin_width = np.diff(bins_hist)
                pdf_arr = bin_mass_hist / bin_width
                X = x[i][j]
                if len(np.where(X < bins_hist)[0]) and len(np.where(X >= bins_hist)[0]):
                    pdf_row.append(pdf_arr[min(np.where(X < bins_hist)[0]) - 1])
                else:
                    pdf_row.append(0)
            pdf.append(pdf_row)
        pdf = np.array(pdf)
        return pdf

    def _log_pdf(self, x):
        """Logarithmic probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            log pdf values at the given points
        """
        bin_mass = self.bin_mass
        bins = self.bins
        lpdf = []

        from warnings import warn

        if self._check_single_array_distr(bins, bin_mass):
            bins = np.array(bins)
            bin_mass = np.array(bin_mass)
            bin_width = np.diff(bins)
            if 0 in bin_mass:
                warn("0 value detected in bin_mass is replaced by a positive const")
                small_value = 1e-100
                bin_mass = np.where(bin_mass == 0, small_value, bin_mass)
            lpdf_arr = np.log(bin_mass / bin_width)
            X = x
            if len(np.where(X < bins)[0]) and len(np.where(X >= bins)[0]):
                lpdf = lpdf_arr[min(np.where(X < bins)[0]) - 1]
            else:
                lpdf = 0
            return lpdf

        x = np.array(x)
        for i in range(len(bins)):
            lpdf_row = []
            for j in range(len(bins[0])):
                X = x[i][j]
                bins_hist = bins[i][j]
                bin_mass_hist = bin_mass[i][j]
                bin_width = np.diff(bins_hist)
                if 0 in bin_mass_hist:
                    warn("0 value detected in bin_mass is replaced by a positive const")
                    small_value = 1e-100
                    bin_mass_hist = np.where(
                        bin_mass_hist == 0, small_value, bin_mass_hist
                    )
                lpdf_arr = np.log(bin_mass_hist / bin_width)
                if len(np.where(X < bins_hist)[0]) and len(np.where(X >= bins_hist)[0]):
                    lpdf_row.append(lpdf_arr[min(np.where(X < bins_hist)[0]) - 1])
                else:
                    lpdf_row.append(0)
            lpdf.append(lpdf_row)
        lpdf = np.array(lpdf)
        return lpdf

    def _cdf(self, x):
        """Cumulative distribution function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the cdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            cdf values at the given points
        """
        bins = self.bins
        bin_mass = self.bin_mass
        cdf = []
        pdf = self._pdf(x)

        if self._check_single_array_distr(bins, bin_mass):
            X = x
            bins_hist = bins
            bin_mass_hist = bin_mass
            cum_sum_mass = np.cumsum(bin_mass_hist)
            # cum_bin_index is an array of all indices
            # of the bins or bin edges that are less than X.
            cum_bin_index = np.where(X >= bins_hist)[0]
            if len(cum_bin_index) == len(bins_hist):
                cdf = 1
            elif len(cum_bin_index) > 1:
                cdf = cum_sum_mass[cum_bin_index[-2]] + pdf * (
                    X - bins_hist[cum_bin_index[-1]]
                )
            elif len(cum_bin_index) == 0:
                cdf = 0
            elif len(cum_bin_index) == 1:
                cdf = pdf * (X - bins_hist[cum_bin_index[-1]])

            return cdf

        x = np.array(x)

        for i in range(len(bins)):
            cdf_row = []
            for j in range(len(bins[0])):
                X = x[i][j]
                bins_hist = bins[i][j]
                bin_mass_hist = bin_mass[i][j]
                cum_sum_mass = np.cumsum(bin_mass_hist)
                # cum_bin_index is an array of all indices
                # of the bins or bin edges that are less than X.
                cum_bin_index = np.where(X >= bins_hist)[0]
                if len(cum_bin_index) == len(bins_hist):
                    cdf_row.append(1)
                elif len(cum_bin_index) > 1:
                    cdf_row.append(
                        cum_sum_mass[cum_bin_index[-2]]
                        + pdf[i][j] * (X - bins_hist[cum_bin_index[-1]])
                    )
                elif len(cum_bin_index) == 0:
                    cdf_row.append(0)
                elif len(cum_bin_index) == 1:
                    cdf_row.append(pdf[i][j] * (X - bins_hist[cum_bin_index[-1]]))
            cdf.append(cdf_row)
        cdf = np.array(cdf)
        return cdf

    def _ppf(self, p):
        """Quantile function = percent point function = inverse cdf.

        Parameters
        ----------
        p : 2D np.ndarray, same shape as ``self``
            values to evaluate the ppf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            ppf values at the given points
        """
        bins = self.bins
        bin_mass = self.bin_mass
        ppf = []
        p = np.array(p)

        if self._check_single_array_distr(bins, bin_mass):
            bins = np.array(bins)
            bin_mass = np.array(bin_mass)
            cum_sum_mass = np.cumsum(bin_mass)
            # print(cum_sum_mass)
            pdf_bins = []
            for bin in bins:
                pdf_bins.append(self._pdf(bin))
            P = p
            cum_bin_index_P = np.where(P >= cum_sum_mass)[0]
            if P < 0 or P > 1:
                X = np.NaN
            elif len(cum_bin_index_P) == 0:
                X = bins[0] + P / pdf_bins[len(cum_bin_index_P)]
            elif len(cum_bin_index_P) > 0:
                if P - cum_sum_mass[cum_bin_index_P[-1]] > 0:
                    X = (
                        bins[cum_bin_index_P[-1] + 1]
                        + (P - cum_sum_mass[cum_bin_index_P[-1]])
                        / pdf_bins[len(cum_bin_index_P)]
                    )
                else:
                    X = bins[cum_bin_index_P[-1] + 1]

            return X

        for i in range(len(bins)):
            ppf_row = []
            for j in range(len(bins[0])):
                P = p[i][j]
                bins_hist = bins[i][j]
                bin_mass_hist = bin_mass[i][j]
                cum_sum_mass = np.cumsum(bin_mass_hist)
                # manually finding pdf of 1D array at all bin edges
                pdf_bins = []
                bin_width = np.diff(bins_hist)
                pdf_arr = bin_mass_hist / bin_width
                for bh in bins_hist:
                    if len(np.where(bh < bins_hist)[0]) and len(
                        np.where(bh >= bins_hist)[0]
                    ):
                        pdf_bins.append(pdf_arr[min(np.where(bh < bins_hist)[0]) - 1])
                    else:
                        pdf_bins.append(0)
                pdf_bins = np.array(pdf_bins)
                # find a way to calculate pdf for 1D array ...
                cum_bin_index_P = np.where(P >= cum_sum_mass)[0]
                if P < 0 or P > 1:
                    ppf_row.append(np.NaN)
                elif len(cum_bin_index_P) == 0:
                    X = bins_hist[0] + P / pdf_bins[len(cum_bin_index_P)]
                    ppf_row.append(round(X, 4))
                elif len(cum_bin_index_P) > 0:
                    if P - cum_sum_mass[cum_bin_index_P[-1]] > 0:
                        X = (
                            bins_hist[cum_bin_index_P[-1] + 1]
                            + (P - cum_sum_mass[cum_bin_index_P[-1]])
                            / pdf_bins[len(cum_bin_index_P)]
                        )
                    else:
                        X = bins_hist[cum_bin_index_P[-1] + 1]
                    ppf_row.append(round(X, 4))
            ppf.append(ppf_row)
        ppf = np.array(ppf)
        return ppf

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {
            "bins": [
                [[0, 1, 2, 3, 4], [5, 5.5, 5.8, 6.5, 7, 7.5]],
                [(2, 12, 5), [0, 1, 2, 3, 4]],
                [[1.5, 2.5, 3.1, 4, 5.4], [-4, -2, -1.5, 5, 10]],
            ],
            "bin_mass": [
                [[0.1, 0.2, 0.3, 0.4], [0.25, 0.1, 0, 0.4, 0.25]],
                [[0.1, 0.2, 0.4, 0.2, 0.1], [0.4, 0.3, 0.2, 0.1]],
                [[0.06, 0.15, 0.09, 0.7], [0.4, 0.15, 0.325, 0.125]],
            ],
            "index": pd.Index(np.arange(3)),
            "columns": pd.Index(np.arange(2)),
        }
        params2 = {
            "bins": [
                [[0, 1.5, 2, 3, 4], [5, 5.5, 5.8, 6.5, 7, 7.5]],
                [(2, 12, 5), [0, 1, 2, 3, 4]],
                [[1.5, 2.5, 3.1, 4, 5.4], [-4, -2, -1.5, 5, 10]],
            ],
            "bin_mass": [
                [[0.1, 0.2, 0.3, 0.4], [0.25, 0.1, 0, 0.4, 0.25]],
                [[0.1, 0.2, 0.4, 0.2, 0.1], [0.4, 0.3, 0.2, 0.1]],
                [[0.06, 0.15, 0.09, 0.7], [0.4, 0.15, 0.325, 0.125]],
            ],
            "index": pd.Index(np.arange(3)),
            "columns": pd.Index(np.arange(2)),
        }

        return [params1, params2]



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Student's t-distribution."""

__author__ = ["Alex-JG3", "ivarzap"]

import numpy as np
import pandas as pd
from scipy.special import betaincinv, gamma, hyp2f1, loggamma

from skpro.distributions.base import BaseDistribution


class TDistribution(BaseDistribution):
    """Student's t-distribution (skpro native).

    Parameters
    ----------
    mu : float or array of float (1D or 2D)
        median of the t-distribution.
        Same as the mean, if it exists.
    sigma : float or array of float (1D or 2D), must be positive
        scale parameter of the t-distribution.
        Same as the standard deviation in the limit of large degrees of freedom.
    df : float or array of float (1D or 2D), must be positive
        Degrees of freedom of the t-distribution.
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions.t import TDistribution

    >>> n = TDistribution(mu=[[0, 1], [2, 3], [4, 5]], sigma=1, df=10)
    """

    _tags = {
        "authors": ["Alex-JG3"],
        "maintainers": ["Alex-JG3"],
        "capabilities:approx": ["pdfnorm", "energy"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf", "ppf"],
        "distr:measuretype": "continuous",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def __init__(self, mu, sigma, df=1, index=None, columns=None):
        self.mu = mu
        self.sigma = sigma
        self.df = df

        super().__init__(index=index, columns=columns)

    def _mean(self):
        """Return expected value of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            expected value of distribution (entry-wise)
        """
        mean_arr = self._bc_params["mu"]
        df = self._bc_params["df"]

        if self.ndim == 0:
            if df <= 1:
                return np.inf
            return mean_arr

        if (df <= 1).any():
            mean_arr = mean_arr.astype(np.float32)
            mean_arr[df <= 1] = np.inf
        return mean_arr

    def _var(self):
        r"""Return element/entry-wise variance of the distribution.

        Let :math:`X` be a random variable with the distribution of `self`.
        Returns,

        .. math::
            \mathbb{V}[X] = \begin{cases}
                \frac{\nu}{\nu - 2} & \text{if} \nu > 2, \\
                \infty              & \text{if} \nu \le 2, \\
            \begin{cases}

        Where :math:`\nu` is the degrees of freedom of the t-distribution.

        Returns
        -------
        pd.DataFrame with same rows, columns as `self`
        variance of distribution (entry-wise)
        """
        sigma = self._bc_params["sigma"]
        df = self._bc_params["df"]
        df_arr = df.astype(np.float32)

        if self.ndim == 0:
            if df <= 2:
                return np.inf
            return sigma**2 * df / (df - 2)

        df_arr[df_arr <= 2] = np.inf
        mask = (df_arr > 2) & (df_arr != np.inf)
        df_arr[mask] = sigma[mask] ** 2 * df_arr[mask] / (df_arr[mask] - 2)
        return df_arr

    def _pdf(self, x):
        """Probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            pdf values at the given points
        """
        df = self._bc_params["df"]
        mu = self._bc_params["mu"]
        sigma = self._bc_params["sigma"]

        pdf_arr = gamma((df + 1) / 2)
        pdf_arr = pdf_arr / (np.sqrt(np.pi * df) * gamma(df / 2))
        pdf_arr = pdf_arr * (1 + ((x - mu) / sigma) ** 2 / df) ** (-(df + 1) / 2)
        pdf_arr = pdf_arr / sigma
        return pdf_arr

    def _log_pdf(self, x):
        """Logarithmic probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            log pdf values at the given points
        """
        df = self._bc_params["df"]
        mu = self._bc_params["mu"]
        sigma = self._bc_params["sigma"]

        lpdf_arr = loggamma((df + 1) / 2)
        lpdf_arr = lpdf_arr - 0.5 * np.log(df * np.pi)
        lpdf_arr = lpdf_arr - loggamma(df / 2)
        lpdf_arr = lpdf_arr - ((df + 1) / 2) * np.log(1 + ((x - mu) / sigma) ** 2 / df)
        lpdf_arr = lpdf_arr - np.log(sigma)
        return lpdf_arr

    def _cdf(self, x):
        """Cumulative distribution function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the cdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            cdf values at the given points
        """
        df = self._bc_params["df"]
        mu = self._bc_params["mu"]
        sigma = self._bc_params["sigma"]

        x_ = (x - mu) / sigma
        cdf_arr = x_ * gamma((df + 1) / 2)
        cdf_arr = cdf_arr * hyp2f1(0.5, (df + 1) / 2, 3 / 2, -(x_**2) / df)
        cdf_arr = 0.5 + cdf_arr / (np.sqrt(np.pi * df) * gamma(df / 2))
        return cdf_arr

    def _ppf(self, p):
        """Quantile function = percent point function = inverse cdf.

        Parameters
        ----------
        p : 2D np.ndarray, same shape as ``self``
            values to evaluate the ppf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            ppf values at the given points
        """
        df = self._bc_params["df"]
        mu = self._bc_params["mu"]
        sigma = self._bc_params["sigma"]

        ppf_arr = p.copy()
        ppf_arr[p == 0.5] = 0.0
        ppf_arr[p <= 0] = -np.inf
        ppf_arr[p >= 1] = np.inf

        mask1 = (p < 0.5) & (p > 0)
        mask2 = (p < 1) & (p > 0.5)
        ppf_arr[mask1] = 1 / betaincinv(0.5 * df[mask1], 0.5, 2 * ppf_arr[mask1])
        ppf_arr[mask2] = 1 / betaincinv(0.5 * df[mask2], 0.5, 2 * (1 - ppf_arr[mask2]))
        ppf_arr[mask1 | mask2] = np.sqrt(ppf_arr[mask1 | mask2] - 1)
        ppf_arr[mask1 | mask2] = np.sqrt(df[mask1 | mask2]) * ppf_arr[mask1 | mask2]
        ppf_arr[mask1] = -ppf_arr[mask1]
        ppf_arr = sigma * ppf_arr + mu
        return ppf_arr

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {
            "mu": [[0, 1], [2, 3], [4, 5]],
            "sigma": 1,
            "df": [2, 3],
        }
        params2 = {
            "mu": 0,
            "sigma": 1,
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),
        }
        # scalar case examples
        params3 = {"mu": -2, "sigma": 3, "df": 4}

        return [params1, params2, params3]



"""Interface adapters to scikit-survival survival forest models."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["fkiraly"]

from skpro.survival.adapters.sksurv import _SksurvAdapter
from skpro.survival.base import BaseSurvReg


class SurvivalForestSkSurv(_SksurvAdapter, BaseSurvReg):
    """Random survival forest from scikit-survival.

    Direct interface to ``sksurv.ensemble.RandomSurvivalForest``, by ``sebp``.

    A random survival forest is a meta estimator that fits a number of
    survival trees on various sub-samples of the dataset and uses
    averaging to improve the predictive accuracy and control over-fitting.
    The sub-sample size is always the same as the original input sample
    size but the samples are drawn with replacement if
    `bootstrap=True` (default).

    In each survival tree, the quality of a split is measured by the
    log-rank splitting rule.

    See [1]_ and [2]_ for further description.

    Parameters
    ----------
    n_estimators : integer, optional, default: 100
        The number of trees in the forest.

    max_depth : int or None, optional, default: None
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_split samples.

    min_samples_split : int, float, optional, default: 6
        The minimum number of samples required to split an internal node:

        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.

    min_samples_leaf : int, float, optional, default: 3
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.

        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.

    min_weight_fraction_leaf : float, optional, default: 0.
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when sample_weight is not provided.

    max_features : int, float, string or None, optional, default: None
        The number of features to consider when looking for the best split:

        - If int, then consider `max_features` features at each split.
        - If float, then `max_features` is a fraction and
          `int(max_features * n_features)` features are considered at each
          split.
        - If "sqrt", then `max_features=sqrt(n_features)`.
        - If "log2", then `max_features=log2(n_features)`.
        - If None, then `max_features=n_features`.

        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.

    max_leaf_nodes : int or None, optional, default: None
        Grow a tree with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    bootstrap : boolean, optional, default: True
        Whether bootstrap samples are used when building trees. If False, the
        whole dataset is used to build each tree.

    oob_score : bool, default: False
        Whether to use out-of-bag samples to estimate
        the generalization accuracy.

    n_jobs : int or None, optional, default: None
        The number of jobs to run in parallel. :meth:`fit`, :meth:`predict`,
        :meth:`decision_path` and :meth:`apply` are all parallelized over the
        trees. ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
        context. ``-1`` means using all processors.

    warm_start : bool, optional, default: False
        When set to ``True``, reuse the solution of the previous call to fit
        and add more estimators to the ensemble, otherwise, just fit a whole
        new forest.

    max_samples : int or float, optional, default: None
        If bootstrap is True, the number of samples to draw from X
        to train each base estimator.

        - If None (default), then draw `X.shape[0]` samples.
        - If int, then draw `max_samples` samples.
        - If float, then draw `max_samples * X.shape[0]` samples. Thus,
          `max_samples` should be in the interval `(0.0, 1.0]`.

    verbose : int, optional, default: 0
        Controls the verbosity when fitting and predicting.

    random_state : int, RandomState instance or None, optional, default: None
        Controls both the randomness of the bootstrapping of the samples used
        when building trees (if ``bootstrap=True``) and the sampling of the
        features to consider when looking for the best split at each node
        (if ``max_features < n_features``).

    Attributes
    ----------
    estimators_ : list of SurvivalTree instances
        The collection of fitted sub-estimators.

    unique_times_ : array of shape = (n_unique_times,)
        Unique time points.

    oob_score_ : float
        Concordance index of the training dataset obtained
        using an out-of-bag estimate.

    References
    ----------
    .. [1] Ishwaran, H., Kogalur, U. B., Blackstone, E. H., & Lauer, M. S. (2008).
           Random survival forests. The Annals of Applied Statistics, 2(3), 841–860.

    .. [2] Ishwaran, H., Kogalur, U. B. (2007). Random survival forests for R.
           R News, 7(2), 25–31. https://cran.r-project.org/doc/Rnews/Rnews_2007-2.pdf.
    """

    _tags = {
        "authors": ["sebp", "fkiraly"],  # sebp credit for interfaced estimator=
    }

    def __init__(
        self,
        n_estimators=100,
        max_depth=None,
        min_samples_split=6,
        min_samples_leaf=3,
        min_weight_fraction_leaf=0.0,
        max_features="sqrt",
        max_leaf_nodes=None,
        bootstrap=True,
        oob_score=False,
        n_jobs=None,
        warm_start=False,
        max_samples=None,
        verbose=0,
        random_state=None,
    ):
        self.n_estimators = n_estimators
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.bootstrap = bootstrap
        self.oob_score = oob_score
        self.n_jobs = n_jobs
        self.warm_start = warm_start
        self.max_samples = max_samples
        self.verbose = verbose
        self.random_state = random_state

        super().__init__()

    def _get_sksurv_class(self):
        """Getter of the sksurv class to be used for the adapter."""
        from sksurv.ensemble.forest import RandomSurvivalForest

        return RandomSurvivalForest

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params1 = {}

        params2 = {
            "n_estimators": 99,
            "max_depth": 5,
            "min_samples_split": 5,
            "min_samples_leaf": 2,
            "max_features": "log2",
            "max_leaf_nodes": 10,
            "oob_score": True,
        }

        return [params1, params2]


class SurvivalForestXtraSkSurv(_SksurvAdapter, BaseSurvReg):
    """Survival random forest with extra randomization-averaging, from scikit-survival.

    Direct interface to ``sksurv.ensemble.ExtraSurvivalTrees``, by ``sebp``.

    This class implements a meta estimator that fits a number of randomized
    survival trees (a.k.a. extra-trees) on various sub-samples of the dataset
    and uses averaging to improve the predictive accuracy and control
    over-fitting. The sub-sample size is always the same as the original
    input sample size but the samples are drawn with replacement if
    ``bootstrap=True`` (default).

    In each randomized survival tree, the quality of a split is measured by
    the log-rank splitting rule.

    Compared to ``RandomSurvivalForest``, randomness goes one step
    further in the way splits are computed. As in
    ``RandomSurvivalForest``, a random subset of candidate features is
    used, but instead of looking for the most discriminative thresholds,
    thresholds are drawn at random for each candidate feature and the best of
    these randomly-generated thresholds is picked as the splitting rule.

    Parameters
    ----------
    n_estimators : integer, optional, default: 100
        The number of trees in the forest.

    max_depth : int or None, optional, default: None
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_split samples.

    min_samples_split : int, float, optional, default: 6
        The minimum number of samples required to split an internal node:

        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.

    min_samples_leaf : int, float, optional, default: 3
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.

        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.

    min_weight_fraction_leaf : float, optional, default: 0.
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when sample_weight is not provided.

    max_features : int, float, string or None, optional, default: None
        The number of features to consider when looking for the best split:

        - If int, then consider `max_features` features at each split.
        - If float, then `max_features` is a fraction and
          `int(max_features * n_features)` features are considered at each
          split.
        - If "sqrt", then `max_features=sqrt(n_features)`.
        - If "log2", then `max_features=log2(n_features)`.
        - If None, then `max_features=n_features`.

        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.

    max_leaf_nodes : int or None, optional, default: None
        Grow a tree with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    bootstrap : boolean, optional, default: True
        Whether bootstrap samples are used when building trees. If False, the
        whole dataset is used to build each tree.

    oob_score : bool, default: False
        Whether to use out-of-bag samples to estimate
        the generalization accuracy.

    n_jobs : int or None, optional, default: None
        The number of jobs to run in parallel. :meth:`fit`, :meth:`predict`,
        :meth:`decision_path` and :meth:`apply` are all parallelized over the
        trees. ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
        context. ``-1`` means using all processors.

    warm_start : bool, optional, default: False
        When set to ``True``, reuse the solution of the previous call to fit
        and add more estimators to the ensemble, otherwise, just fit a whole
        new forest.

    max_samples : int or float, optional, default: None
        If bootstrap is True, the number of samples to draw from X
        to train each base estimator.

        - If None (default), then draw `X.shape[0]` samples.
        - If int, then draw `max_samples` samples.
        - If float, then draw `max_samples * X.shape[0]` samples. Thus,
          `max_samples` should be in the interval `(0.0, 1.0]`.

    verbose : int, optional, default: 0
        Controls the verbosity when fitting and predicting.

    random_state : int, RandomState instance or None, optional, default: None
        Controls both the randomness of the bootstrapping of the samples used
        when building trees (if ``bootstrap=True``) and the sampling of the
        features to consider when looking for the best split at each node
        (if ``max_features < n_features``).

    Attributes
    ----------
    estimators_ : list of SurvivalTree instances
        The collection of fitted sub-estimators.

    unique_times_ : array of shape = (n_unique_times,)
        Unique time points.

    oob_score_ : float
        Concordance index of the training dataset obtained
        using an out-of-bag estimate.
    """

    _tags = {
        "authors": ["sebp", "fkiraly"],  # sebp credit for interfaced estimator=
    }

    def __init__(
        self,
        n_estimators=100,
        max_depth=None,
        min_samples_split=6,
        min_samples_leaf=3,
        min_weight_fraction_leaf=0.0,
        max_features="sqrt",
        max_leaf_nodes=None,
        bootstrap=True,
        oob_score=False,
        n_jobs=None,
        warm_start=False,
        max_samples=None,
        verbose=0,
        random_state=None,
    ):
        self.n_estimators = n_estimators
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.bootstrap = bootstrap
        self.oob_score = oob_score
        self.n_jobs = n_jobs
        self.warm_start = warm_start
        self.max_samples = max_samples
        self.verbose = verbose
        self.random_state = random_state

        super().__init__()

    def _get_sksurv_class(self):
        """Getter of the sksurv class to be used for the adapter."""
        from sksurv.ensemble.forest import ExtraSurvivalTrees

        return ExtraSurvivalTrees

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params1 = {}

        params2 = {
            "n_estimators": 99,
            "max_depth": 5,
            "min_samples_split": 5,
            "min_samples_leaf": 2,
            "max_features": "log2",
            "max_leaf_nodes": 10,
            "oob_score": True,
        }

        return [params1, params2]



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Module for nested equality checking."""
from skpro.utils.deep_equals._deep_equals import deep_equals

__all__ = [
    "deep_equals",
]



"""Tests for tag register an tag functionality."""

from skpro.registry._tags import OBJECT_TAG_REGISTER


def test_tag_register_type():
    """Test the specification of the tag register. See _tags for specs."""
    assert isinstance(OBJECT_TAG_REGISTER, list)
    assert all(isinstance(tag, tuple) for tag in OBJECT_TAG_REGISTER)

    for tag in OBJECT_TAG_REGISTER:
        assert len(tag) == 4
        assert isinstance(tag[0], str)
        assert isinstance(tag[1], (str, list))
        if isinstance(tag[1], list):
            assert all(isinstance(x, str) for x in tag[1])
        assert isinstance(tag[2], (str, tuple))
        if isinstance(tag[2], tuple):
            assert len(tag[2]) == 2
            assert isinstance(tag[2][0], str)
            assert isinstance(tag[2][1], (list, str))
            if isinstance(tag[2][1], list):
                assert all(isinstance(x, str) for x in tag[2][1])
        assert isinstance(tag[3], str)



"""Test functionality of time series plotting functions."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

import pytest
from skbase.utils.dependencies import _check_soft_dependencies

from skpro.tests.test_switch import run_test_module_changed


@pytest.mark.skipif(
    not run_test_module_changed("skpro.utils")
    or not _check_soft_dependencies("matplotlib", severity="none"),
    reason="skip test if required soft dependency for matplotlib not available",
)
def test_plot_crossplot_interval():
    """Test that plot_crossplot_interval runs without error."""
    _check_soft_dependencies("matplotlib")

    from sklearn.datasets import load_diabetes
    from sklearn.ensemble import RandomForestRegressor
    from sklearn.linear_model import LinearRegression

    from skpro.regression.residual import ResidualDouble
    from skpro.utils.plotting import plot_crossplot_interval

    X, y = load_diabetes(return_X_y=True, as_frame=True)
    reg_mean = LinearRegression()
    reg_resid = RandomForestRegressor()
    reg_proba = ResidualDouble(reg_mean, reg_resid)

    reg_proba.fit(X, y)
    y_pred_proba = reg_proba.predict_proba(X)

    plot_crossplot_interval(y, y_pred_proba, coverage=0.8)
    plot_crossplot_interval(y, y_pred_proba)

    y_pred_interval = reg_proba.predict_interval(X, coverage=0.7)
    plot_crossplot_interval(y, y_pred_interval)


@pytest.mark.skipif(
    not run_test_module_changed("skpro.utils")
    or not _check_soft_dependencies("matplotlib", severity="none"),
    reason="skip test if required soft dependency for matplotlib not available",
)
def test_plot_crossplot_std():
    """Test that plot_crossplot_std runs without error."""
    _check_soft_dependencies("matplotlib")

    from sklearn.datasets import load_diabetes
    from sklearn.ensemble import RandomForestRegressor
    from sklearn.linear_model import LinearRegression

    from skpro.regression.residual import ResidualDouble
    from skpro.utils.plotting import plot_crossplot_std

    X, y = load_diabetes(return_X_y=True, as_frame=True)
    reg_mean = LinearRegression()
    reg_resid = RandomForestRegressor()
    reg_proba = ResidualDouble(reg_mean, reg_resid)

    reg_proba.fit(X, y)
    y_pred = reg_proba.predict_proba(X)

    plot_crossplot_std(y, y_pred)

    y_pred_var = reg_proba.predict_var(X)
    plot_crossplot_std(y, y_pred_var)


@pytest.mark.skipif(
    not run_test_module_changed("skpro.utils")
    or not _check_soft_dependencies("matplotlib", severity="none"),
    reason="skip test if required soft dependency for matplotlib not available",
)
def test_plot_crossplot_loss():
    """Test that plot_crossplot_loss runs without error."""
    _check_soft_dependencies("matplotlib")

    from sklearn.datasets import load_diabetes
    from sklearn.ensemble import RandomForestRegressor
    from sklearn.linear_model import LinearRegression

    from skpro.metrics import CRPS
    from skpro.regression.residual import ResidualDouble
    from skpro.utils.plotting import plot_crossplot_loss

    X, y = load_diabetes(return_X_y=True, as_frame=True)
    reg_mean = LinearRegression()
    reg_resid = RandomForestRegressor()
    reg_proba = ResidualDouble(reg_mean, reg_resid)

    reg_proba.fit(X, y)
    y_pred = reg_proba.predict_proba(X)

    crps_metric = CRPS()
    plot_crossplot_loss(y, y_pred, crps_metric)



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Erlang probability distribution."""

__author__ = ["RUPESH-KUMAR01"]

import pandas as pd
from scipy.stats import erlang

from skpro.distributions.adapters.scipy import _ScipyAdapter


class Erlang(_ScipyAdapter):
    r"""Erlang Distribution.

    Most methods wrap ``scipy.stats.erlang``.

    The Erlang Distribution is parameterized by shape :math:`k`
    and rate :math:`\lambda`, such that the pdf is

    .. math:: f(x) = \frac{x^{k-1}\exp\left(-\lambda x\right) \lambda^{k}}{(k-1)!}

    Parameters
    ----------
    rate : float or array of float (1D or 2D)
        Represents the rate parameter, which is also the inverse of the scale parameter.
    k : int or array of int (1D or 2D), optional, default = 1
        Represents the shape parameter.
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions.erlang import Erlang

    >>> d = Erlang(rate=[[1, 1], [2, 3], [4, 5]], k=2)
    """

    _tags = {
        "capabilities:approx": ["energy", "pdfnorm"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf", "ppf"],
        "distr:measuretype": "continuous",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def __init__(self, rate, k=1, index=None, columns=None):
        self.rate = rate
        self.k = k

        super().__init__(index=index, columns=columns)

    def _get_scipy_object(self):
        return erlang

    def _get_scipy_param(self):
        rate = self._bc_params["rate"]
        k = self._bc_params["k"]

        return [], {"scale": 1 / rate, "a": k}

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # Array case examples
        params1 = {
            "rate": 2.0,
            "k": 3,
            "index": pd.Index([0, 1, 2]),
            "columns": pd.Index(["x", "y"]),
        }
        # Scalar case examples
        params2 = {"rate": 0.8, "k": 2}

        params3 = {"rate": 3.0, "k": 1}

        return [params1, params2, params3]
